/* $Id: address.c,v 1.17 2000/01/15 22:29:02 chrisf Exp $ */

/*
Hot Date - A DatebookDB displayer for the PalmPilot
Copyright (C) 1999 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
 * Just so you know these routines are pulled from the Address source code
 * as provided by Palm.  At least that is what I started from, I may have
 * touched a few lines to get it to compile with gcc without warnings.
 */

#include <Pilot.h>
#include "callback.h"
#include "hotdateRsc.h"
#include "hotdate.h"
#include "address.h"

extern struct sPrefsR *PrefsR;
extern DmOpenRef AddressDB;
extern ULong Prefscontactslaunch;
extern Boolean PrefsRdirty;
extern Boolean SortByCompany;

char *UnnamedRecordStringPtr="";

/*
 * Static function prototypes
 */
static Boolean FindAddressApp(ULong AppID, UInt *cardNo, LocalID *dbID);
static Boolean DetermineRecordName(AddrDBRecordPtr recordP, 
    UInt *shortenedFieldWidth, UInt *fieldSeparatorWidth, Boolean sortByCompany,
    CharPtr *name1, Int *name1Length, Int *name1Width, 
    CharPtr *name2, Int *name2Length, Int *name2Width,
    CharPtr *unnamedRecordStringPtr, UInt nameExtent);
static void DrawRecordName(
    CharPtr name1, Int name1Length, Int name1Width, 
    CharPtr name2, Int name2Length, Int name2Width,
    UInt nameExtent, SWord *x, SWord y, UInt shortenedFieldWidth, 
    UInt fieldSeparatorWidth, Boolean center, Boolean priorityIsName1);
static void DrawRecordNameAndPhoneNumber(AddrDBRecordPtr record, 
    RectanglePtr bounds, Boolean sortByCompany, CharPtr *unnamedRecordStringPtr);
static void ContactsPopupDrawItem(UInt itemNum, RectanglePtr bounds,
    CharPtr *unusedP);

/* Extract the bit at position index from bitfield.  0 is the high bit. */
#define BitAtPosition(pos)                ((ULong)1 << (pos))
#define GetBitMacro(bitfield, index)      ((bitfield) & BitAtPosition(index))
#define SetBitMacro(bitfield, index)      ((bitfield) |= BitAtPosition(index))
#define RemoveBitMacro(bitfield, index)   ((bitfield) &= ~BitAtPosition(index))

/************************************************************
 *
 *  FUNCTION: AddrUnpack
 *
 *  DESCRIPTION: Fills in the AddrDBRecord structure
 *
 *  PARAMETERS: address record to unpack
 *                the address record to unpack into
 *
 *  RETURNS: the record unpacked
 *
 *  CREATED: 1/14/95 
 *
 *  BY: Roger Flores
 *
 *************************************************************/
void AddrUnpack(AddrPackedDBRecord *src, AddrDBRecordPtr dest)
{
    Int index;
    ULong flags;
    char *p;

    dest->options = src->options;
    flags = src->flags.allBits;
    p = &src->firstField;
         
    for (index=firstAddressField; index < addressFieldsCount; index++) {
        /* If the flag is set point to the string else NULL */
        if (GetBitMacro(flags, index) != 0) {
            dest->fields[index] = p;
            p += StrLen(p) + 1;
        } else dest->fields[index] = NULL;
    }
}

static Boolean FindAddressApp(ULong AppID, UInt *cardNo, LocalID *dbID)
{
    DmSearchStateType searchInfo;

    /*
     * Grab the id of the application we want to launch.  This is
     * defined by preferences.  Who knows what'll happen if the
     * launched program doesn't like the structure.  ToDo seems
     * to be especially unstable if launched with a Datebook
     * record.  That's because they don't verify if the record
     * number is correct, and they assume it is always within
     * ToDoDB!  I'll assign the punishment -- castration!
     *
     * Hrm, looking at both Datebook and ToDo source it appears that
     * there is just no error testing at all to control a condition
     * where the record number passed is out of range.  So what am
     * I supposed to do!?!  I don't want the device to fatal error
     * just because the user selected an inappropriate application
     * to launch.
     *
     * The only thing I can do is limit the list to known working
     * applications!  God what a horrible situation.
     */

    if ((AppID == 'addr') ||            /* built-in Address */
        (AppID == 'addp') ||            /* Address+ */
        (AppID == 'JZa1') ||            /* AddressPro */
        (AppID == 'SNms') ||            /* Super Names */
        (AppID == 'TlPh')) {            /* TealPhone (2.20 & higher) */
        if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
            sysFileTApplication, AppID, true, cardNo, dbID) == 0)) return true;
    }
    return false;
}

/***********************************************************************
 *
 * FUNCTION:    DetermineRecordName
 *
 * DESCRIPTION: Determines an address book record's name.  The name
 * varies based on which fields exist and what the sort order is.
 *
 * PARAMETERS:  name1, name2 - first and seconds names to draw
 *              name1Length, name2Length - length of the names in chars
 *              name1Width, name2Width - width of the names when drawn
 *              nameExtent - the space the names must be drawn in
 *              *x, y - where the names are drawn
 *              shortenedFieldWidth - the width in the current font
 *              
 *
 * RETURNED:    x is set after the last char drawn
 *                     Boolean - name1/name2 priority based on sortByCompany
 *
 * REVISION HISTORY:
 *            Name        Date        Description
 *            ----        ----        -----------
 *            roger        6/20/95    Initial Revision
 *            frigino    970813    Added priority return value
 *
 ***********************************************************************/
 
static Boolean DetermineRecordName(AddrDBRecordPtr recordP, 
    UInt *shortenedFieldWidth, UInt *fieldSeparatorWidth, Boolean sortByCompany,
    CharPtr *name1, Int *name1Length, Int *name1Width, 
    CharPtr *name2, Int *name2Length, Int *name2Width,
    CharPtr *unnamedRecordStringPtr, UInt nameExtent)
{
    UInt fieldNameChoiceList[4];
    UInt fieldNameChoice;
    Boolean ignored;
    Boolean name1HasPriority;

    *shortenedFieldWidth =  FntCharsWidth(shortenedFieldString, 
        shortenedFieldLength);
    *fieldSeparatorWidth = FntCharsWidth(fieldSeparatorString, 
        fieldSeparatorLength);

    *name1 = NULL;
    *name2 = NULL;

    if (sortByCompany) {
        /* When sorting by company, always treat name2 as priority. */
        name1HasPriority = false;
        fieldNameChoiceList[3] = addressFieldsCount;
        fieldNameChoiceList[2] = firstName;
        fieldNameChoiceList[1] = name;
        fieldNameChoiceList[0] = company;
        fieldNameChoice = 0;

        while ((*name1 == NULL) &&
            (fieldNameChoiceList[fieldNameChoice] != addressFieldsCount)) {
            *name1 = recordP->fields[fieldNameChoiceList[fieldNameChoice++]];
        }

        /*
         * When sorting by company, treat name2 as priority if we
         * succeed in getting the company name as the name1
         * Did we get the company name?
         */
        if (fieldNameChoice > 1) {
            /*
             * No. We got a last name, first name, or nothing.
             * Priority switches to name1
             */
            name1HasPriority = true;
        }

        while ((*name2 == NULL) &&
            (fieldNameChoiceList[fieldNameChoice] != addressFieldsCount)) {
            *name2 = recordP->fields[fieldNameChoiceList[fieldNameChoice++]];
        }
    } else {
        /* When not sorting by company, always treat name1 as priority. */
        name1HasPriority = true;

        fieldNameChoiceList[3] = addressFieldsCount;
        fieldNameChoiceList[2] = addressFieldsCount;
        fieldNameChoiceList[1] = firstName;
        fieldNameChoiceList[0] = name;
        fieldNameChoice = 0;

        while ((*name1 == NULL) &&
            (fieldNameChoiceList[fieldNameChoice] != addressFieldsCount)) {
            *name1 = recordP->fields[fieldNameChoiceList[fieldNameChoice++]];
        }

        if (*name1 == NULL) {
            *name1 = recordP->fields[company];
            *name2 = NULL;
        } else {
            while ((*name2 == NULL) &&
                (fieldNameChoiceList[fieldNameChoice] != addressFieldsCount)) {
                *name2 =
                    recordP->fields[fieldNameChoiceList[fieldNameChoice++]];
            }
        }         
    }

    if (*name1) {
        /* Only show text from the first line in the field */
        *name1Length = nameExtent; /* longer than possible */
        *name1Width = nameExtent;  /* wider than possible */
        FntCharsInWidth(*name1, name1Width, name1Length, &ignored);
    } else {
        /* Set the name to the unnamed string */
        *name1 = *unnamedRecordStringPtr;
        *name1Length = StrLen(*unnamedRecordStringPtr);
        *name1Width = FntCharsWidth(*unnamedRecordStringPtr, *name1Length);
    }
      
    if (*name2) {
        /* Only show text from the first line in the field */
        *name2Length = nameExtent; /* longer than possible */
        *name2Width = nameExtent;  /* wider than possible */
        FntCharsInWidth(*name2, name2Width, name2Length, &ignored);
    } else {
        *name2Length = 0;
        *name2Width = 0;
    }

    /* Return priority status */
    return name1HasPriority;
}

/***********************************************************************
 *
 * FUNCTION:    DrawRecordName
 *
 * DESCRIPTION: Draws an address book record name.  It is used
 * for the list view and note view.
 *
 * PARAMETERS:  name1, name2 - first and seconds names to draw
 *              name1Length, name2Length - length of the names in chars
 *              name1Width, name2Width - width of the names when drawn
 *              nameExtent - the space the names must be drawn in
 *              *x, y - where the names are drawn
 *              shortenedFieldWidth - the width in the current font
 *              
 *
 * RETURNED:    x is set after the last char drawn
 *
 * REVISION HISTORY:
 *            Name        Date        Description
 *            ----        ----        -----------
 *            roger        6/20/95    Initial Revision
 *            frigino    970813    Rewritten. Now includes a variable ratio for
 *                                    name1/name2 width allocation, a prioritization
 *                                    parameter, and a word break search to allow
 *                                    reclaiming of space from the low priority name.
 *
 ***********************************************************************/

static void DrawRecordName(
    CharPtr name1, Int name1Length, Int name1Width, 
    CharPtr name2, Int name2Length, Int name2Width,
    UInt nameExtent, SWord *x, SWord y, UInt shortenedFieldWidth, 
    UInt fieldSeparatorWidth, Boolean center, Boolean priorityIsName1)
{
    Int name1MaxWidth;
    Int name2MaxWidth;
    Boolean ignored;
    Int totalWidth;
    CharPtr lowPriName;
    Int highPriNameWidth;
    Int lowPriNameWidth;
    Int highPriMaxWidth;
    Int lowPriMaxWidth;
    CharPtr spaceP;

    /* Check if both names fit */
    totalWidth = name1Width + (name2 ? fieldSeparatorWidth : 0) + name2Width;

    /*
     * If we are supposed to center the names then move in the x position
     * by the amount that centers the text
     */
    if (center && (nameExtent > totalWidth)) {
        *x += (nameExtent - totalWidth) / 2;
    }

    /* Special case if only name1 is given */
    if (name2 == NULL) {
        /*
         * For some reason, OS3 changed the code so that it doesn't show
         * ellipses if there is only name1.  I liked it the old way!
         */
        /* Does name1 fit in its maximum width? */
        if (name1Width > nameExtent) {
            /* No. Draw it to max width minus the ellipsis */
            name1Width = nameExtent-shortenedFieldWidth;
            FntCharsInWidth(name1, &name1Width, &name1Length, &ignored);
            WinDrawChars(name1, name1Length, *x, y);
            *x += name1Width;
            /* Draw ellipsis */
            WinDrawChars(shortenedFieldString, shortenedFieldLength, *x, y);
            *x += shortenedFieldWidth;
        } else {
            /* Yes. Draw name1 within its width */
            FntCharsInWidth(name1, &name1Width, &name1Length, &ignored);
            WinDrawChars(name1, name1Length, *x, y);
            *x += name1Width;
        }
        return;
    }

    /* Remove name separator width */
    nameExtent -= fieldSeparatorWidth;

    /* Test if both names fit */
    if ((name1Width + name2Width) <= nameExtent) {
        name1MaxWidth = name1Width;
        name2MaxWidth = name2Width;
    } else {
        /*
         * They dont fit. One or both needs truncation
         * Establish name priorities and their allowed widths
         * Change this to alter the ratio of the low and high
         * priority name spaces
         */
        highPriMaxWidth = (nameExtent << 1)/3; /* 1/3 to low and 2/3 to high */
        lowPriMaxWidth = nameExtent-highPriMaxWidth;

        /* Save working copies of names and widths based on priority */
        if (priorityIsName1) {
            /* Priority is name1 */
            highPriNameWidth = name1Width;
            lowPriName = name2;
            lowPriNameWidth = name2Width;
        } else {
            /* Priority is name2 */
            highPriNameWidth = name2Width;
            lowPriName = name1;
            lowPriNameWidth = name1Width;
        }

        /* Does high priority name fit in high priority max width? */
        if (highPriNameWidth > highPriMaxWidth) {
            /* No. Look for word break in low priority name */
            spaceP = StrChr(lowPriName, spaceChr);
            if (spaceP != NULL) {
                /* Found break. Set low priority name width to break width */
                lowPriNameWidth = FntCharsWidth(lowPriName, spaceP-lowPriName);
                /*
                 * Reclaim width from low pri name width to low pri max width,
                 * if smaller.
                 */
                if (lowPriNameWidth < lowPriMaxWidth) {
                    lowPriMaxWidth = lowPriNameWidth;
                    /* Set new high pri max width */
                    highPriMaxWidth = nameExtent-lowPriMaxWidth;
                }
            }
        } else {
            /* Yes. Adjust maximum widths */
            highPriMaxWidth = highPriNameWidth;
            lowPriMaxWidth = nameExtent-highPriMaxWidth;
        }

        /* Convert priority widths back to name widths */
        if (priorityIsName1) {
            /* Priority is name1 */
            name1Width = highPriNameWidth;
            name2Width = lowPriNameWidth;
            name1MaxWidth = highPriMaxWidth;
            name2MaxWidth = lowPriMaxWidth;
        } else {
            /* Priority is name2 */
            name1Width = lowPriNameWidth;
            name2Width = highPriNameWidth;
            name1MaxWidth = lowPriMaxWidth;
            name2MaxWidth = highPriMaxWidth;
        }
    }

    /* Does name1 fit in its maximum width? */
    if (name1Width > name1MaxWidth) {
        /* No. Draw it to max width minus the ellipsis */
        name1Width = name1MaxWidth-shortenedFieldWidth;
        FntCharsInWidth(name1, &name1Width, &name1Length, &ignored);
        WinDrawChars(name1, name1Length, *x, y);
        *x += name1Width;

        /* Draw ellipsis */
        WinDrawChars(shortenedFieldString, shortenedFieldLength, *x, y);
        *x += shortenedFieldWidth;
    } else {
        /* Yes. Draw name1 within its width */
        FntCharsInWidth(name1, &name1Width, &name1Length, &ignored);
        WinDrawChars(name1, name1Length, *x, y);
        *x += name1Width;
    }

    /* Draw name separator */
    WinDrawChars(fieldSeparatorString, fieldSeparatorLength, *x, y);
    *x += fieldSeparatorWidth;

    /* Draw name2 within its maximum width */
    FntCharsInWidth(name2, &name2MaxWidth, &name2Length, &ignored);
    WinDrawChars(name2, name2Length, *x, y);
    *x += name2MaxWidth;
}

/***********************************************************************
 *
 * FUNCTION:    DrawRecordNameAndPhoneNumber
 *
 * DESCRIPTION: Draws the namd and phone number (plus which phone)
 * within the screen bounds passed.
 *
 * PARAMETERS:  record - record to draw
 *              bounds - bounds of the draw region
 *              phoneLabelLetters - the first letter of each phone label
 *              sortByCompany - true if the database is sorted by company
 *              unnamedRecordStringPtr - string to use for unnamed records
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *         Name   Date      Description
 *         ----   ----      -----------
 *         roger   6/21/95   Initial Revision
 *
 ***********************************************************************/
 
static void DrawRecordNameAndPhoneNumber(AddrDBRecordPtr record, 
    RectanglePtr bounds, Boolean sortByCompany, CharPtr *unnamedRecordStringPtr)
{
    SWord x, y;
    UInt phoneLabel;
    UInt fieldSeparatorWidth;
    UInt shortenedFieldWidth;
    CharPtr name1, name2, phone;
    Int name1Length;
    Int name2Length;
    Int phoneLength;
    Int name1Width;
    Int name2Width;
    Int phoneWidth;
    UInt nameExtent;
    Boolean ignored;
    Boolean name1HasPriority;
    short phoneLabelWidth;
    char *phoneLabelLetters ="WHFOEMPM";

    x = bounds->topLeft.x;
    y = bounds->topLeft.y;

    /* remove the blank trailing column */
    phoneLabelWidth = FntCharWidth('W') - 1;

    bounds->extent.x -= (phoneLabelWidth + 1);

    name1HasPriority = DetermineRecordName(record, &shortenedFieldWidth,
        &fieldSeparatorWidth, sortByCompany, &name1, &name1Length, &name1Width,
        &name2, &name2Length, &name2Width, unnamedRecordStringPtr,
        bounds->extent.x);

    phone = record->fields[phone1 + record->options.phones.displayPhoneForList];
    if (phone) {
        /* Only show text from the first line in the field */
        phoneWidth = bounds->extent.x;
        phoneLength = StrLen(phone);
        FntCharsInWidth(phone, &phoneWidth, &phoneLength, &ignored);
    } else {
        phoneLength = 0;
        phoneWidth = 0;
    }

    if (bounds->extent.x >= name1Width+(name2 ? fieldSeparatorWidth : 0) +
        name2Width+(phone ? spaceBetweenNamesAndPhoneNumbers : 0)+phoneWidth) {
        /* we can draw it all! */
        WinDrawChars(name1, name1Length, x, y);
        x += name1Width;

        /* Is there a second name? */
        if (name2) {
            if (name1) {
                WinDrawChars(fieldSeparatorString, fieldSeparatorLength, x, y);
                x += fieldSeparatorWidth;
            }
            /* draw name2 */
            WinDrawChars(name2, name2Length, x, y);
            x += name2Width;
        }

        if (phone)
            WinDrawChars(phone, phoneLength,
                bounds->topLeft.x + bounds->extent.x - phoneWidth, y);
    } else {
        /* Shortened math (970812 maf) */
        nameExtent = bounds->extent.x - MIN(phoneWidth, PhoneColumnWidth);

        /*
         * Leave some space between names and numbers if there is a
         * phone number.
         */
        if (phone) nameExtent -= spaceBetweenNamesAndPhoneNumbers;

        DrawRecordName(name1, name1Length, name1Width, name2, name2Length,
            name2Width, nameExtent, &x, y, shortenedFieldWidth,
            fieldSeparatorWidth, false, name1HasPriority || !sortByCompany);

        if (phone) {
            x += spaceBetweenNamesAndPhoneNumbers;
            nameExtent = x - bounds->topLeft.x;

            /* Now draw the phone number */
            if (bounds->extent.x - nameExtent >= phoneWidth) {
                /* We can draw it all */
                WinDrawChars(phone, phoneLength,
                    bounds->topLeft.x + bounds->extent.x - phoneWidth, y);
            } else {
                /*
                 * The phone number should be right justified instead of using
                 * x from above because the string printed may be shorter
                 * than we expect (CharsInWidth chops off space chars).
                 */
                phoneWidth = bounds->extent.x-nameExtent-shortenedFieldWidth;
                FntCharsInWidth(phone, &phoneWidth, &phoneLength, &ignored);
                WinDrawChars(phone, phoneLength,
                    bounds->topLeft.x+bounds->extent.x-
                    shortenedFieldWidth-phoneWidth, y);
                WinDrawChars(shortenedFieldString, shortenedFieldLength, 
                    bounds->topLeft.x+bounds->extent.x-shortenedFieldWidth, y);
            }
        }
    }

    if (phone) {
        /* Draw the first letter of the phone field label */
        phoneLabel = GetPhoneLabel(record, firstPhoneField +
            record->options.phones.displayPhoneForList);
        WinDrawChars(&phoneLabelLetters[phoneLabel], 1,
            bounds->topLeft.x + bounds->extent.x + 1 + ((phoneLabelWidth -
            (FntCharWidth(phoneLabelLetters[phoneLabel]) - 1)) >> 1), y);
    }
}

static void ContactsPopupDrawItem(UInt itemNum, RectanglePtr bounds,
    CharPtr *unusedP)
{
    SWord tl;
    AddrPackedDBRecord *rp;
    AddrDBRecordType r;
    UInt rindex;
    VoidHand recordH=0;
    ULong uniqueid;
    VoidHand strH;
    CharPtr strP;
    RectangleType b;

    CALLBACK_PROLOGUE

    if (itemNum >= PrefsR->Contacts.listcount) {
        if ((strH = DmGetResource(strRsc, strID_editcontacts))) {
            strP = MemHandleLock(strH);
            tl = StrLen(strP);
            if (tl)
                WinDrawChars(strP, tl, bounds->topLeft.x, bounds->topLeft.y);
            MemHandleUnlock(strH);
            DmReleaseResource(strH);
        }
    } else {
        MemMove(&b, bounds, sizeof(b));
        /* fudge factor for the scroll arrows */
        b.extent.x -= PrefsR->Contacts.listcount > 10 ? 9 : 2;
        MemMove(&uniqueid, &PrefsR->Contacts.list[itemNum*sizeof(ULong)],
            sizeof(ULong));
        if (DmFindRecordByID(AddressDB, uniqueid, &rindex) == 0) {
            if ((recordH = DmQueryRecord(AddressDB, rindex))) {
                rp = (AddrPackedDBRecord *) MemHandleLock(recordH);
                /*
                 * Build the unpacked structure for an AddressDB record.  It
                 * is just a bunch of pointers into the rp structure.
                 */
                AddrUnpack(rp, &r);
                DrawRecordNameAndPhoneNumber(&r, &b, SortByCompany,
                    &UnnamedRecordStringPtr);
                MemHandleUnlock(recordH);
            } else {
                if ((strH = DmGetResource(strRsc, strID_deleted))) {
                    strP = MemHandleLock(strH);
                    tl = StrLen(strP);
                    if (tl) WinDrawChars(strP, tl, bounds->topLeft.x,
                        bounds->topLeft.y);
                    MemHandleUnlock(strH);
                    DmReleaseResource(strH);
                }
            }
        } else {
            if ((strH = DmGetResource(strRsc, strID_notfound))) {
                strP = MemHandleLock(strH);
                tl = StrLen(strP);
                if (tl) WinDrawChars(strP, tl, bounds->topLeft.x,
                    bounds->topLeft.y);
                MemHandleUnlock(strH);
                DmReleaseResource(strH);
            }
        }
    }

    CALLBACK_EPILOGUE
}

Int ContactsPopup(Word listid)
{
    Int curSelection=-1;
    Int newSelection;
    ListPtr lst;
    FormPtr frm;
    short height;
    RectangleType r;
    UInt cardNo, AcardNo;
    LocalID dbID, AdbID;
    DmSearchStateType searchInfo;
    GoToParamsPtr theGotoPointer;
    UInt rindex;
    ULong uniqueid;
    ULong AppID, DataID;
    UInt attr;
    
    frm = FrmGetActiveForm();
    lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listid));
    LstSetDrawFunction(lst, (ListDrawDataFuncPtr) ContactsPopupDrawItem);

    height = (unsigned char) PrefsR->Contacts.listcount & 0xff;
    height++;
    LstSetListChoices(lst, NULL, height);
    LstSetHeight(lst, MIN(height, 11));
    FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, listid), &r);
    LstSetPosition(lst, r.topLeft.x, 134-r.extent.y);
    LstSetSelection(lst, curSelection);
    LstSetTopItem(lst, 0);

    newSelection = LstPopupList(lst);

    if ((newSelection != curSelection) && (newSelection != -1)) {

        /* Edit Contacts */
        if (newSelection == height-1) {
            FrmPopupForm(EditContactsForm);
        } else {

        /* picked a contract, so launch the app */
        do {

        DataID = AddressAppID;
        AppID = Prefscontactslaunch;
        if (!FindAddressApp(AppID, &AcardNo, &AdbID)) break;

        /* Allocate the new pointer for the goto parameters. */
        theGotoPointer = MemPtrNew(sizeof(GoToParamsType));
        if (!theGotoPointer) break;
        /*
         * Set the owner of the pointer to be the system. This is
         * required because all memory allocated by our application
         * is freed when the application quits. Our application will
         * quit the next time through our event handler.
         */
        if ((MemPtrSetOwner(theGotoPointer, 0) == 0) &&
            (DmGetNextDatabaseByTypeCreator(true, &searchInfo,
            'DATA', DataID, true, &cardNo, &dbID) == 0)) {

            MemMove(&uniqueid,
                &PrefsR->Contacts.list[newSelection*sizeof(ULong)],
                sizeof(ULong));
            if (DmFindRecordByID(AddressDB, uniqueid, &rindex) != 0) break;
            if (DmRecordInfo(AddressDB, rindex, &attr, NULL, NULL) != 0) break;
            /*
             * Make sure to check if the record is deleted first.  Otherwise
             * the Address App will choke on the goto.
             */
            if (attr & dmRecAttrDelete) break;

            /*
             * Copy all the goto information into the GoToParamsPtr
             * structure.
             */
            theGotoPointer->searchStrLen = 0;
            theGotoPointer->dbCardNo = cardNo;
            theGotoPointer->dbID = dbID;
            theGotoPointer->recordNum = rindex;
            theGotoPointer->matchPos = 0;
            theGotoPointer->matchFieldNum = 0;
            theGotoPointer->matchCustom = 0;
            SysUIAppSwitch(AcardNo, AdbID, sysAppLaunchCmdGoTo,
                (Ptr) theGotoPointer);
        } else {
            MemPtrFree(theGotoPointer);
        }

        } while (0);
        }
    }
    
    return newSelection;
}

Boolean EditContactsHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    ListPtr lst;
    Int height;
    UInt cardNo;
    LocalID dbID;
    ULong AppID;
    DWord result;
    Word selection;
    ULong tempitem;
    AddrLookupParamsType addrlookup;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        lst = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_editcontactslist));
        LstSetDrawFunction(lst, (ListDrawDataFuncPtr) ContactsPopupDrawItem);
        height = (unsigned char) PrefsR->Contacts.listcount & 0xff;
        LstSetListChoices(lst, NULL, height);
        LstSetSelection(lst, -1);
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_editcontactsok:
            FrmReturnToForm(MainForm);
            handled = true;
            break;
        case buttonID_editcontactsadd:
            /*
             * TealPoint was nice enough to add the sysAppLaunchCmdGoTo
             * but we found that adding sysAppLaunchCmdLookup was too
             * much of a burden since it would just duplicate the built-in
             * Address Book lookup.  So we will use the built-in one.  I'm
             * doing the same thing with Address+ because it's Lookup is
             * buggy.
             *
             * Actually, you know what.. I am gonna just do that for all
             * of them because I haven't found an Address Book replacement
             * without a buggy lookup feature.
             */
            AppID = AddressAppID;

            if (!FindAddressApp(AppID, &cardNo, &dbID)) break;

            addrlookup.title = NULL;
            addrlookup.pasteButtonText = NULL;
            addrlookup.lookupString[0] = (char) NULL;
            addrlookup.field1 = addrLookupSortField;
            addrlookup.field2 = addrLookupListPhone;
            addrlookup.field2Optional = true;
            addrlookup.userShouldInteract = true;
            addrlookup.formatStringP = NULL;
            if ((SysAppLaunch(cardNo, dbID, 0, sysAppLaunchCmdLookup,
                (Ptr)&addrlookup, &result) != 0) ||
                (addrlookup.uniqueID == 0))
                break;

            /*
             * Add the record's uniqueID to the end of our contacts list.
             */
            height = (unsigned char) PrefsR->Contacts.listcount & 0xff;
            if (height < MAXCONTACTS) {
                MemMove(&PrefsR->Contacts.list[height*sizeof(ULong)],
                    &addrlookup.uniqueID, sizeof(ULong));
                (unsigned char) PrefsR->Contacts.listcount = ++height;
                PrefsRdirty = true;
                frm = FrmGetActiveForm();
                lst = FrmGetObjectPtr(frm,
                    FrmGetObjectIndex(frm, listID_editcontactslist));
                LstSetListChoices(lst, NULL, height);
                LstSetSelection(lst, height-1);
                LstSetTopItem(lst, height-1);
                LstDrawList(lst);
            }
            handled = true;
            break;
        case buttonID_editcontactsdelete:
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, listID_editcontactslist));
            height = (unsigned char) PrefsR->Contacts.listcount & 0xff;
            if (((selection = LstGetSelection(lst)) != -1) &&
                (selection < height)) {
                /* scroll the contacts list up */
                if (selection < height-1) {
                    MemMove(&PrefsR->Contacts.list[(selection)*sizeof(ULong)],
                        &PrefsR->Contacts.list[(selection+1)*sizeof(ULong)],
                        ((height-(selection+1))*sizeof(ULong)));
                }
                (unsigned char) PrefsR->Contacts.listcount = --height;
                PrefsRdirty = true;
                frm = FrmGetActiveForm();
                lst = FrmGetObjectPtr(frm,
                    FrmGetObjectIndex(frm, listID_editcontactslist));
                LstSetListChoices(lst, NULL, height);
                if (selection >= height) selection = height-1;
                LstSetSelection(lst, selection);
                LstDrawList(lst);
            }
            handled = true;
            break;
        case buttonID_editcontactsup:
        case buttonID_editcontactsdown:
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, listID_editcontactslist));
            height = (unsigned char) PrefsR->Contacts.listcount & 0xff;
            if ((selection = LstGetSelection(lst)) != -1) {
                if (e->data.ctlSelect.controlID == buttonID_editcontactsup) {
                    /* swap the current item with the preceeding one */
                    if (selection > 0) {
                        MemMove(&tempitem,
                            &PrefsR->Contacts.list[(selection)*sizeof(ULong)],
                            sizeof(ULong));
                        MemMove(
                            &PrefsR->Contacts.list[(selection)*sizeof(ULong)],
                            &PrefsR->Contacts.list[(selection-1)*sizeof(ULong)],
                            sizeof(ULong));
                        MemMove(
                            &PrefsR->Contacts.list[(selection-1)*sizeof(ULong)],
                            &tempitem,
                            sizeof(ULong));
                        selection--;
                    } else break;
                } else {
                    /* swap the current item with the following one */
                    if (selection < height-1) {
                        MemMove(&tempitem,
                            &PrefsR->Contacts.list[(selection)*sizeof(ULong)],
                            sizeof(ULong));
                        MemMove(
                            &PrefsR->Contacts.list[(selection)*sizeof(ULong)],
                            &PrefsR->Contacts.list[(selection+1)*sizeof(ULong)],
                            sizeof(ULong));
                        MemMove(
                            &PrefsR->Contacts.list[(selection+1)*sizeof(ULong)],
                            &tempitem,
                            sizeof(ULong));
                        selection++;
                    } else break;
                }
                PrefsRdirty = true;
                frm = FrmGetActiveForm();
                lst = FrmGetObjectPtr(frm,
                    FrmGetObjectIndex(frm, listID_editcontactslist));
                LstSetListChoices(lst, NULL, height);
                LstSetSelection(lst, selection);
                LstDrawList(lst);
            }
            handled = true;
            break;
        }
        break;

    case keyDownEvent:
        if (e->data.keyDown.chr == pageUpChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_editcontactslist));
            LstScrollList(lst, up, LstGetVisibleItems(lst));
        } else if (e->data.keyDown.chr == pageDownChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_editcontactslist));
            LstScrollList(lst, down, LstGetVisibleItems(lst));
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}
