/* $Id: hotdate.c,v 1.278 2000/05/26 00:56:13 chrisf Exp $ */

/*
Hot Date - A DatebookDB displayer for the PalmPilot
Copyright (C) 1999 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <Pilot.h>
#include <System/DataPrv.h>
#include "callback.h"
#include "hotdateRsc.h"
#include "hotdate.h"
#include "datebook.h"
#include "todo.h"
#include "address.h"
#include "timer.h"

/* globals */

Char AppErrStr[AppErrStrLen];
Boolean PrefsRdirty, PrefsWeredirty, ignorependown;
DateType CurrentDate;
Word tps;
Word AppsPopupCount;
Handle AppsPopupH;
Boolean inposer;
Boolean SortByCompany=true;
SystemPreferencesType sysPrefs;

char *pageupchr="\001", *pagedownchr="\002";

/*
 * FME - May 14, 2000: No longer needed because of change in ShowScrollArrows.
 * char *pageupgreychr="\003", *pagedowngreychr="\004";
 */

/*
 * This is for the Week tab's label which I set to the starting day when
 * it is pressed.
 */
char todaylabel[]="12/31/99", weeklabel[]="12/31";
char *origtodaylabel, *origweeklabel;

/*
 * This is the list of appointments, or really just handles to allocated
 * memory.  The list is created in MainFormLoadTable() for the week, and
 * also if the device is turned on and the day has changed.  When I am
 * drawing the table I use the recordNum member to grab the text.
 */
#define MAXDAYS 31
UInt apptCounts[MAXDAYS];
VoidHand apptLists[MAXDAYS];
ApptInfoType apptCurrent;
Boolean apptCurrentNote;
UInt apptCurrentRow;
UInt apptTop, apptTotal, apptnumRows, apptDays, apptAlarmLine, apptToDoLine;
Word apptToDoCategory;
DateType apptDate;
UInt Prefsdays;
short Prefsdfmt1, Prefsdfmt2, Prefstfmt;
ULong Prefsapptlaunch, Prefstodolaunch, Prefslowerleftlaunch,
    Prefscontactslaunch;
ULong Prefshardlaunch[4];
short lasttab=-1;
short apptTickMark;
DWord dbcdate, dbmdate, tdcdate, tdmdate;

/*
 * These are the times that correspond with the list selections for the
 * pizza timer.
 */
#define TIMESINLIST 18
UInt times[]={0, 30, 60, 60*2, 60*3, 60*4, 60*5, 60*6, 60*7, 60*8, 60*9,
    60*10, 60*15, 60*20, 60*25, 60*30, 60*45, 60*60};

/*
 * Prefs stuff
 */
VoidHand PrefsRecHandle, PrefsRHandle;
UInt PrefsRecIndex;
struct sPrefsR *PrefsR;
struct sPrefsR DefaultPrefsR={
    PREFSVERSION,
    { '0', '0', '0', '1', '1', '0', 'H', '0', '1', '0', '0' },
    { '0', '0', '0', '0', '1', '1', '0' },
    { "date", "todo", "date", "addr", { '0', '0', '0', '0' },
      { "date", "addr", "todo", "memo" } },
    "All",
    { "\0", "\0\0\0", "\0\0\0", "\0\0\0", "\0\0\0", "\0", "\0", "\0", "" },
    { (char) 0, "" }
};

/*
 * These are for the private database
 */
DmOpenRef MainDB;

/*
 * These are for accessing the built-in DatebookDB
 */
DmOpenRef DatebookDB;

/*
 * These are for accessing the built-in ToDoDB
 */
DmOpenRef ToDoDB;

/*
 * These are for accessing the built-in AddressDB
 */
DmOpenRef AddressDB;

/*
 * Static function prototypes
 */
static void MigratePrefs(struct sPrefsR *currP, struct s126PrefsR *oldP);
static void MainFormDrawRecord(VoidPtr tableP, Word row, Word column,
    RectanglePtr bounds);
static void ShowTickMark(FormPtr frm, Boolean eraseonly);
static Boolean MainFormHandleEvent(EventPtr e);
static void DateToString(char *str, DateType d, TimeType t1, TimeType t2);
static Boolean NotesFormHandleEvent(EventPtr e);
static Boolean ApptPrefsFormHandleEvent(EventPtr e);
static Boolean ToDoPrefsFormHandleEvent(EventPtr e);
static Boolean LaunchPrefsFormHandleEvent(EventPtr e);
static Boolean AboutFormHandleEvent(EventPtr e);
static Boolean ContribFormHandleEvent(EventPtr e);
static Boolean ApplicationHandleEvent(EventPtr e);
static void freelistitems(void);
static void freememories(void);
static Word StartApplication(void);
static void StopApplication(void);
static void refreshlist(short new_tab);
static void EventLoop(void);
static void AppsPopupDrawItem(UInt itemNum, RectanglePtr bounds,
    CharPtr *unusedP);
static ULong AppsPopup(Word listid, Word triggerid, ULong curcreator);
static void CheckPrefsR(void);
static void ReadPrefsRec(void);
static void WritePrefsRec(void);
static short FindDateFormat1(char format);
static short FindDateFormat2(char format);
static short FindTimeFormat(char format);
static void LoadApptPrefsFields(void);
static void UnloadApptPrefsFields(Boolean save);
static void LoadToDoPrefsFields(void);
static void UnloadToDoPrefsFields(Boolean save);
static void LoadLaunchPrefsFields(void);
static void UnloadLaunchPrefsFields(Boolean save);

/* Added by FME - May 6, 2000. Will be called from MainFormHandleEvent */
static void UpdateWeekLabel(FormPtr frm, DateType BeginDate);

static void MigratePrefs(struct sPrefsR *currP, struct s126PrefsR *oldP)
{
    /* first initialize a new structure to the defaults */
    MemMove(currP, &DefaultPrefsR, sizeof(DefaultPrefsR));
    /* then copy over stuff from the older prefs, where applicable */
    MemMove(&currP->ApptPrefs, &oldP->ApptPrefs, sizeof(currP->ApptPrefs));
    MemMove(&currP->ToDoPrefs, &oldP->ToDoPrefs, sizeof(currP->ToDoPrefs));
    /*
     * Don't forget that I added a contactslaunch, so I have to do these
     * individually and let contactslaunch take the default from above.
     */
    MemMove(&currP->LaunchPrefs.apptlaunch, &oldP->LaunchPrefs.apptlaunch,
        sizeof(currP->LaunchPrefs.apptlaunch));
    MemMove(&currP->LaunchPrefs.todolaunch, &oldP->LaunchPrefs.todolaunch,
        sizeof(currP->LaunchPrefs.todolaunch));
    MemMove(&currP->LaunchPrefs.lowerleftlaunch,
        &oldP->LaunchPrefs.lowerleftlaunch,
        sizeof(currP->LaunchPrefs.lowerleftlaunch));
    MemMove(&currP->LaunchPrefs.hardlaunchflag,
        &oldP->LaunchPrefs.hardlaunchflag,
        sizeof(currP->LaunchPrefs.hardlaunchflag));
    MemMove(&currP->LaunchPrefs.hardlaunch, &oldP->LaunchPrefs.hardlaunch,
        sizeof(currP->LaunchPrefs.hardlaunch));
}

static void MainFormDrawRecord(VoidPtr tableP, Word row, Word column,
    RectanglePtr bounds)
{
    ApptPackedDBRecordPtr apptrecP;
    ToDoDBRecord *todorecP;
    VoidHand recHandle;
    VoidHand strH;
    CharPtr strP;
    UInt apptnum, rownum;
    LineItemPtr li;
    long l1, l2;
    short day;
    Int textfit, width, width2, datew=0, timew=0;
    Int x, y;
    Boolean fits;
    FontID currFont;
    short i1;

    CALLBACK_PROLOGUE

    /*
     * Get the appointment number that corresponds to the table item to draw.
     * The appointment number is stored in the RowID field of the row.
     */
    rownum = TblGetRowID(tableP, row);
    apptnum = rownum;

    /*
     * See which day (and in turn apptLists) we need to start at.
     */
    for (day=0; day < apptDays; day++) {
        if (apptnum < apptCounts[day]) break;
        apptnum -= apptCounts[day];
    }

    if ((day >= apptDays) ||
        !(li = MemHandleLock(apptLists[day]))) {
        return;
    }

    /*
     * Set the standard font.  Save the current font.
     * It is a Pilot convention to not destroy the current font
     * but to save and restore it.
     */
    currFont = FntSetFont(stdFont);

    /*
     * The title for untimed alarms.  Right now I use the same code to draw
     * the untimed event items as the normal datebook items.
     */

    if (rownum == apptAlarmLine) {

        FntSetFont(boldFont);
        if ((strH = DmGetResource(strRsc, strID_untimedtitle))) {
            strP = MemHandleLock(strH);
            i1 = StrLen(strP);
            width = FntCharsWidth(strP, i1);
            WinDrawChars(strP, i1, bounds->topLeft.x+10, bounds->topLeft.y);
            MemHandleUnlock(strH);
            DmReleaseResource(strH);
        } else width = 0;

        WinDrawLine(bounds->topLeft.x, bounds->topLeft.y+(bounds->extent.y/2),
            bounds->topLeft.x+7, bounds->topLeft.y+(bounds->extent.y/2));
        WinDrawLine(bounds->topLeft.x+width+12,
            bounds->topLeft.y+(bounds->extent.y/2),
            bounds->topLeft.x+bounds->extent.x-1,
            bounds->topLeft.y+(bounds->extent.y/2));
        FntSetFont(stdFont);

    /*
     * I use a global apptToDoLine to indicate the row number of the ToDo
     * title.  The ToDo items will follow.  Everything before that is a
     * Datebook appointment.
     */

    } else if (rownum == apptToDoLine) {

        /* Title for todo */
        FntSetFont(boldFont);
        if ((strH = DmGetResource(strRsc, strID_todotitle))) {
            strP = MemHandleLock(strH);
            i1 = StrLen(strP);
            width = FntCharsWidth(strP, i1);
            WinDrawChars(strP, i1, bounds->topLeft.x+10,
                bounds->topLeft.y);
            MemHandleUnlock(strH);
            DmReleaseResource(strH);
        } else width = 0;

        WinDrawLine(bounds->topLeft.x, bounds->topLeft.y+(bounds->extent.y/2),
            bounds->topLeft.x+7, bounds->topLeft.y+(bounds->extent.y/2));
        /*
         * Draw the category name on this line as well.
         */
        FntSetFont(stdFont);
        i1 = StrLen(PrefsR->todocategory);
        width2 = FntCharsWidth(PrefsR->todocategory, i1);
        WinDrawLine(bounds->topLeft.x+width+12,
            bounds->topLeft.y+(bounds->extent.y/2),
            bounds->topLeft.x+bounds->extent.x-width2-15,
            bounds->topLeft.y+(bounds->extent.y/2));
        WinDrawChars(PrefsR->todocategory, i1, bounds->topLeft.x+
            bounds->extent.x-width2-11, bounds->topLeft.y);
        /* now draw the triangle graphic */
        x = bounds->topLeft.x+bounds->extent.x-8;
        y = bounds->topLeft.y+4;
        WinDrawLine(x, y, x+7, y);
        WinDrawLine(x+1, y+1, x+6, y+1);
        WinDrawLine(x+2, y+2, x+5, y+2);
        WinDrawLine(x+3, y+3, x+4, y+3);

    } else if (rownum > apptToDoLine) {

        /* This is a todo */
        if ((recHandle = DmQueryRecord(ToDoDB, li[apptnum].recordNum))) {
            todorecP = MemHandleLock(recHandle);

            if (DateToInt(todorecP->dueDate) != toDoNoDueDate) {
                DateToAscii(todorecP->dueDate.month, todorecP->dueDate.day,
                    todorecP->dueDate.year+1904, Prefsdfmt2, AppErrStr);
                /* slash, dot, or dash */
                switch (Prefsdfmt2) {
                case dfYMDWithDots:
                case dfYMDWithDashes:
                case dfYMDWithSlashes:
                    WinDrawChars(&AppErrStr[3], StrLen(AppErrStr)-3,
                        bounds->topLeft.x, bounds->topLeft.y);
                    break;
                default:
                    WinDrawChars(AppErrStr, StrLen(AppErrStr)-3, bounds->topLeft.x,
                        bounds->topLeft.y);
                    break;
                }
                /* Is it overdue? */
                if (DateToInt(todorecP->dueDate) < DateToInt(apptDate)) {
                    width = FntCharsWidth(AppErrStr, StrLen(AppErrStr)-3);
                    FntSetFont(boldFont);
                    WinDrawChars("!", 1, bounds->topLeft.x+width,
                        bounds->topLeft.y);
                    FntSetFont(stdFont);
                }
                datew = 29;
            } else {
                /* get rid of the date column for undated todos */
                datew = 0;
            }

            /* Calculate the position of the description */
            l1 = sizeof(DateType)+sizeof(unsigned char);
            /* Draw the description */
            textfit = StrLen(&todorecP->description);
            width2 = 0;
            if ((MemHandleSize(recHandle)-(l1+(textfit+1))) > 1) {
                /*
                 * There is a note on this record, but don't draw the
                 * symbol if not set in the prefs.
                 */
                if (PrefsR->ApptPrefs.symnote == '1') {
                    FntSetFont(symbolFont);
                    width2 = FntCharWidth('\013');
                    FntSetFont(stdFont);
                }
                /*
                 * I test this when the user clicks to display the note.
                 */
                TblSetItemInt(tableP, row, 0, 1);
            }
            width = bounds->extent.x-datew-width2-1;
            FntCharsInWidth(&todorecP->description, &width, &textfit, &fits);
            WinDrawChars(&todorecP->description, textfit,
                bounds->topLeft.x+datew, bounds->topLeft.y);
            /* Put a little note icon on the right */
            if (width2) {
                FntSetFont(symbolFont);
                WinDrawChars("\013", 1, bounds->topLeft.x+bounds->extent.x-
                    width2, bounds->topLeft.y);
                FntSetFont(stdFont);
            }

            /* Unlock the handle to the record */
            MemHandleUnlock(recHandle);
        }

    } else {

        /* This is a datebook appointment */
        if ((recHandle = DmQueryRecord(DatebookDB,
            li[apptnum].recordNum))) {
            apptrecP = MemHandleLock(recHandle);
            /*
             * Show the date on week view if prefs say so.  Also show it
             * in the untimed list.
             */
            if (((apptDays > 1) && (PrefsR->ApptPrefs.showdate == '1')) ||
                ((apptDays == 1) && (apptnum > apptAlarmLine))) {
                DateToAscii(li[apptnum].date.month, li[apptnum].date.day,
                    li[apptnum].date.year+1904, Prefsdfmt2,
                    AppErrStr);
                /* slash, dot, or dash */
                switch (Prefsdfmt2) {
                case dfYMDWithDots:
                case dfYMDWithDashes:
                case dfYMDWithSlashes:
                    WinDrawChars(&AppErrStr[3], StrLen(AppErrStr)-3,
                        bounds->topLeft.x, bounds->topLeft.y);
                    break;
                default:
                    WinDrawChars(AppErrStr, StrLen(AppErrStr)-3,
                        bounds->topLeft.x, bounds->topLeft.y);
                    break;
                }
                datew = 27;
            } else datew = 0;
            /* The day of the week */
            DateToDOWDMFormat(li[apptnum].date.month, li[apptnum].date.day,
                li[apptnum].date.year+1904, dfYMDLongWithSpace, AppErrStr);
            WinDrawChars(AppErrStr, 3, bounds->topLeft.x+datew,
                bounds->topLeft.y);
            datew += 19;

            /* The starting time */
            if (TimeToInt(apptrecP->when.startTime) != apptNoTime) {
                TimeToAscii(apptrecP->when.startTime.hours,
                    apptrecP->when.startTime.minutes, Prefstfmt, AppErrStr);
                l1 = StrLen(AppErrStr);
                if (l1 >= 4) {
                    /* move the a or p in a little closer, if it exists */
                    if ((UCASE(AppErrStr[l1-2]) == 'A') ||
                        (UCASE(AppErrStr[l1-2]) == 'P')) {
                        AppErrStr[l1-3] = AppErrStr[l1-2];
                        l1 -= 2;
                        width2 = 28;
                    } else {
                        /* there's a little extra room without am/pm */
                        width2 = 26;
                    }
                    /* let's right-justify the time */
                    width = (width2)-FntCharsWidth(AppErrStr, l1);
                    if (width < 0) width = 0;
                        WinDrawChars(AppErrStr, l1,
                            bounds->topLeft.x+datew+width, bounds->topLeft.y);
                }
                timew = 30;
            } else {
                /*
                 * We don't want the diamond in the untimed list.
                 */
                if (!(rownum > apptAlarmLine)) {
                    /* draw the little diamond to indicate an untimed event */
                    FntSetFont(symbolFont);
                    width = (14)-(FntCharWidth('\023')/2);
                    if (width < 0) width = 0;
                    WinDrawChars("\023", 1, bounds->topLeft.x+datew+width,
                        bounds->topLeft.y);
                    FntSetFont(stdFont);
                    timew = 30;
                } else timew = 3;
            }

            /* Calculate the position of the description */
            l1 = sizeof(ApptDateTimeType)+sizeof(ApptDBRecordFlags);
            if (apptrecP->flags.alarm) l1 += sizeof(AlarmInfoType);
            if (apptrecP->flags.repeat) l1 += sizeof(RepeatInfoType);
            if (apptrecP->flags.exceptions)
                l1 += sizeof(UInt) +
                    (((ExceptionsListType *)
                    ((CharPtr) apptrecP+l1))->numExceptions*sizeof(DateType));

            width2 = -1; /* keeps the rightmost icon flush */

            /* grab the length of the description here */
            if (apptrecP->flags.description)
                textfit = StrLen((CharPtr) apptrecP+l1);
            else textfit = 0;

            /*
             * Put a little note icon on the right.  This is a little different
             * than the Datebook position because we want to always let the
             * user tap on the far-right to launch a note.
             */
            if (apptrecP->flags.note) {
                /*
                 * Datebk3 uses the notes field for flags.  Let's ignore the
                 * note if all it contains is these flags.
                 */
                l2 = l1+textfit+1;
                if ((StrLen((CharPtr) apptrecP+l2) == 10) &&
                    (*((CharPtr) apptrecP+l2) == '#') &&
                    (*((CharPtr) apptrecP+l2+1) == '#') &&
                    (*((CharPtr) apptrecP+l2+9) == '\n')) {
                    /* note is just Datebk3 flags, therefore ignore it */
                } else {
                    if (PrefsR->ApptPrefs.symnote == '1') {
                        if (width2 <= 0) FntSetFont(symbolFont);
                        width2 += FntCharWidth('\013')+1;
                        WinDrawChars("\013", 1,
                            bounds->topLeft.x+bounds->extent.x-
                            width2, bounds->topLeft.y);
                    }
                    /*
                     * I test this when the user clicks to display the note.
                     */
                    TblSetItemInt(tableP, row, 0, 1);
                }
            }

            /* Put a little repeat icon on the right */
            if ((PrefsR->ApptPrefs.symrepeat == '1') &&
                apptrecP->flags.repeat) {
                if (width2 <= 0) FntSetFont(symbolFont);
                width2 += FntCharWidth('\025')+1;
                WinDrawChars("\025", 1, bounds->topLeft.x+bounds->extent.x-
                    width2, bounds->topLeft.y);
            }

            /* Put a little alarm icon on the right */
            if ((PrefsR->ApptPrefs.symalarm == '1') && apptrecP->flags.alarm) {
                if (width2 <= 0) FntSetFont(symbolFont);
                width2 += FntCharWidth('\024')+1;
                WinDrawChars("\024", 1, bounds->topLeft.x+bounds->extent.x-
                    width2, bounds->topLeft.y);
            }

            if (width2 == -1) width2 = 0;
            if (width2) FntSetFont(stdFont);

            /* Draw the description */
            if (apptrecP->flags.description) {
                width = bounds->extent.x-(datew+timew)-width2;
                /* textfit is already grabbed above */
                FntCharsInWidth((CharPtr) apptrecP+l1, &width, &textfit,
                    &fits);
                WinDrawChars((CharPtr) apptrecP+l1, textfit,
                    bounds->topLeft.x+(datew+timew), bounds->topLeft.y);
            }

            /* Unlock the handle to the record */
            MemHandleUnlock(recHandle);
        }

        /* at the bottom of every day draw a gray line */
        if ((row != (apptnumRows-1)) && (apptnum == (apptCounts[day]-1)) &&
            (rownum != (apptAlarmLine-1)) && (rownum != (apptToDoLine-1)) &&
            (apptDays > 1)) {
            WinDrawGrayLine(bounds->topLeft.x,
                bounds->topLeft.y+bounds->extent.y-1,
                bounds->topLeft.x+bounds->extent.x-1,
                bounds->topLeft.y+bounds->extent.y-1);
        }

    }

    /* Restore the font */
    FntSetFont (currFont);

    MemPtrUnlock(li);

    CALLBACK_EPILOGUE
}

void MainFormLoadTable(FormPtr frm, short redraw, UInt top, DateType date,
    UInt days)
{
    Word row, refreshrow;
    TablePtr tableP;
    short i;
    UInt apptnum, tcnt;
    Boolean selectable, needtocache=false;
    char *ptr;
    DateType tdate;

    if (days == 0) return;

    apptTickMark = -1;
    ShowTickMark(frm, true); /* erase the current tick mark */
    tableP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, tableID_dates));
    apptnumRows = TblGetNumberOfRows(tableP);
    refreshrow = 0; /* row to start refreshing at, mostly for quick ToDo */

    /*
     * If top is -1 then we'll read the appointments from the database.  And
     * initialize most of the globals for the table.
     */
    if (redraw == 1) {
        TblEraseTable(tableP);
        apptTop = top;
        apptDays = days;
        apptDate = date;
        apptAlarmLine = 65535;
        apptToDoLine = 65535;
        /*
         * Clear up all appointments
         */
        freelistitems();

        /*
         * Here we are looking to speed things up by pulling the line items
         * from our stored preferences record.  This way, when we start the
         * program it will display the initial list very quickly as long as
         * none of the databases has changed.
         */
        tcnt = (UInt) (((unsigned char) PrefsR->DayCache.listcount[0] << 8)+
            (unsigned char) PrefsR->DayCache.listcount[1]);
        MemMove(&tdate, PrefsR->DayCache.date, sizeof(DateType));
        /*
         * I don't do the caching for the week view yet.  And make sure we
         * don't use the item count if it is greater than apptMaxPerDay.  This
         * is because of a bug in one of the released versions.
         */
        if ((tcnt > 0) && (tcnt <= apptMaxPerDay) && (days == 1) &&
            (DateToInt(tdate) == DateToInt(date)) &&
            ((short) (PrefsR->ApptPrefs.defaulttab-'0') == lasttab) &&
            (MemCmp((char *) &dbcdate, PrefsR->DayCache.dbcdate, 4) == 0) &&
            (MemCmp((char *) &dbmdate, PrefsR->DayCache.dbmdate, 4) == 0) &&
            (MemCmp((char *) &tdcdate, PrefsR->DayCache.tdcdate, 4) == 0) &&
            (MemCmp((char *) &tdmdate, PrefsR->DayCache.tdmdate, 4) == 0)) {
            if (tcnt > 0) {
                apptLists[0] =
                    MemHandleNew(tcnt*sizeof(LineItemType));
                if (apptLists[0] && (ptr = MemHandleLock(apptLists[0]))) {
                    MemMove(ptr, PrefsR->DayCache.list,
                        tcnt*sizeof(LineItemType));
                    apptCounts[0] = tcnt;
                    apptAlarmLine = (UInt)
                        (((unsigned char) PrefsR->DayCache.alarmline[0] << 8)+
                        (unsigned char) PrefsR->DayCache.alarmline[1]);
                    apptToDoLine = (UInt)
                         ((unsigned char) (PrefsR->DayCache.todoline[0] << 8)+
                        (unsigned char) PrefsR->DayCache.todoline[1]);
                    apptTotal = apptCounts[0];
                    MemPtrUnlock(ptr);
                } else return;
            }
        } else {

            /*
             * Do the real loading if it isn't available in the cache.
             */
            apptTotal = ApptGetAppointments(DatebookDB, date, apptDays,
                apptLists, apptCounts);
            apptTotal += ApptGetUntimedAlarms(DatebookDB, date, apptDays,
                apptLists, apptCounts);
            /* Only do the ToDos if the preferences are set */
            if (PrefsR->ToDoPrefs.range != '0') {
                apptTotal += GetToDo(ToDoDB, date, apptDays, apptLists,
                    apptCounts);
            }
            needtocache = true;

        }

    }

    /*
     * Here is a special redraw case that I use when I cycle through the
     * ToDo categories.  I only want to regenerate the ToDo items and not
     * have to wait for the appt and untimed stuff.  Since the ToDos are
     * at the end, I'll subtract them from the count and get them again.
     */
    if (redraw == 3) {
        if (apptDays >= 1) {
            apptCounts[apptDays-1] -= apptTotal-apptToDoLine;
            apptTotal = apptToDoLine;
            apptTotal += GetToDo(ToDoDB, date, apptDays, apptLists, apptCounts);
            refreshrow = apptToDoLine-apptTop;
            needtocache = true;
            /*
             * Let's invalidate the cache because the user probably cycled
             * the ToDo category, in which case the cache will be invalid
             * even if he is not on the default tab.
             */
            PrefsR->DayCache.listcount[0] = (char) 0;
            PrefsR->DayCache.listcount[1] = (char) 0;
            PrefsRdirty = true;
        }
    }

    /*
     * Let's write the line items out to our preferences record.  We
     * don't actually store it in the record yet, but we set a flag so
     * it gets saved when we exit.
     */
    if ((days == 1) && needtocache &&
        ((short) (PrefsR->ApptPrefs.defaulttab-'0') == lasttab)) {
        MemMove(PrefsR->DayCache.date, &date, sizeof(DateType));
        MemMove(PrefsR->DayCache.dbcdate, (char *) &dbcdate, 4);
        MemMove(PrefsR->DayCache.dbmdate, (char *) &dbmdate, 4);
        MemMove(PrefsR->DayCache.tdcdate, (char *) &tdcdate, 4);
        MemMove(PrefsR->DayCache.tdmdate, (char *) &tdmdate, 4);
        if ((apptCounts[0] > 0) && apptLists[0] &&
            (ptr = MemHandleLock(apptLists[0]))) {
            MemMove(PrefsR->DayCache.list, ptr,
                apptCounts[0]*sizeof(LineItemType));
            MemPtrUnlock(ptr);
        }
        PrefsR->DayCache.listcount[0] = apptCounts[0] >> 8;
        PrefsR->DayCache.listcount[1] = apptCounts[0] & 0xff;
        PrefsR->DayCache.alarmline[0] = apptAlarmLine >> 8;
        PrefsR->DayCache.alarmline[1] = apptAlarmLine & 0xff;
        PrefsR->DayCache.todoline[0] = apptToDoLine >> 8;
        PrefsR->DayCache.todoline[1] = apptToDoLine & 0xff;
        PrefsRdirty = true;
    }

    /*
     * Get the appointment that will show at the top.
     */
    apptnum = top;
    for (i=0; i < apptDays; i++) {
        if (apptnum < apptCounts[i]) break;
        apptnum -= apptCounts[i];
    }

    if (i >= apptDays) return;

    for (row=0; row < apptnumRows; row++) {
        /*
         * Shift to the next day's appointments when necessary.
         */
        while ((i < apptDays) && (apptnum >= apptCounts[i])) {
            apptnum -= apptCounts[i++];
        }
        if ((i < apptDays) && (apptnum < apptCounts[i])) {
            /*
             * Set customTableItem to indicate that we want to be called
             * to draw the record.
             */
            TblSetItemStyle(tableP, row, 0, customTableItem);
            /*
             * The ItemInt is used for flags.  Currently I am only interested
             * in whether the row has a note.  I don't feel like digging into
             * the record here to determine that, since it is already done
             * in MainFormDrawRecord().  So I set it to 0 to start with.
             */
            TblSetItemInt(tableP, row, 0, 0);
            /* The RowID I set to the appointment number */
            TblSetRowID(tableP, row, row+top);
            TblSetRowHeight(tableP, row, FntCharHeight());
            TblSetRowUsable(tableP, row, true);
            selectable = true;
            TblSetRowSelectable(tableP, row, selectable);
        } else {
            /*
             * Sometimes there are more table rows than appointments.
             * If this happens mark those unused table rows as not usable.
             */
            TblSetRowUsable(tableP, row, false);
        }
        /*
         * Mark the row invalid so that it will draw when we call the
         * draw routine.  The refreshrow is used so that a partial redraw
         * can be done in the case of a ToDo category change.
         */
        if (row >= refreshrow) TblMarkRowInvalid(tableP, row);
        apptnum++;
    }

    /* Set the column usable so that it draws and accepts user input. */
    TblSetColumnUsable(tableP, 0, true);

    /* Set the callback routine that will draw the records. */
    TblSetCustomDrawProcedure(tableP, 0, MainFormDrawRecord);

    if (redraw == 1) TblDrawTable(tableP);
    else if (redraw >= 2) {
        TblUnhighlightSelection(tableP);
        TblRedrawTable(tableP);
    }
}

/*
 * Just some grunt work to show/hide the pageup/pagedown arrows.  I have
 * two buttons and I replace their text with the appropriate arrow thingy
 * from the Symbol 7 font.
 *
 * FME - May 14, 2000: Replaced with more understandable code using the
 *                     standard function FrmUpdateScrollers.
 */
void ShowScrollArrows(FormPtr frm)
{
    Word    upIndex;
    Word    downIndex;
    Boolean scrollUp;
    Boolean scrollDown;

    scrollUp = (apptTop != 0);
    scrollDown = ((apptTop+apptnumRows) < apptTotal);

    upIndex = FrmGetObjectIndex(frm, buttonID_pageup);
    downIndex = FrmGetObjectIndex(frm, buttonID_pagedown);

    FrmUpdateScrollers(frm, upIndex, downIndex, scrollUp, scrollDown);
}

/*
 * Show the tick mark next to the closest appointment.  Only call this when
 * today is showing.
 */
static void ShowTickMark(FormPtr frm, Boolean eraseonly)
{
    TablePtr tableP;
    TimeType CurrentTime;
    DateTimeType dt;
    short i, howmany, newtick;
    LineItemPtr li;
    RectangleType r;

    TimSecondsToDateTime(TimGetSeconds(), &dt);
    CurrentTime.hours = dt.hour;
    CurrentTime.minutes = dt.minute;

    tableP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, tableID_dates));
    TblGetBounds(tableP, &r);
    if (eraseonly) {
        WinEraseLine(0, r.topLeft.y, 0, r.topLeft.y+r.extent.y-1);
        apptTickMark = -1;
        return;
    }

    /* make sure it is only for today */
    if ((apptDate.day != dt.day) || (apptDate.month != dt.month) ||
        ((apptDate.year+1904) != dt.year)) return;

    if (!apptLists[0] || !(li = MemHandleLock(apptLists[0]))) return;
    howmany = MIN(TblGetNumberOfRows(tableP), apptCounts[0]-apptTop);
    /* don't consider any items after appointments */
    if ((howmany+apptTop) >= apptAlarmLine)
        howmany -= (howmany+apptTop)-apptAlarmLine;
    if ((howmany+apptTop) >= apptToDoLine)
        howmany -= (howmany+apptTop)-apptToDoLine;
    newtick = -1;
    for (i=0; i < howmany; i++) {
        if (TimeToInt(CurrentTime) < TimeToInt(li[apptTop+i].startTime)) break;
        newtick = i;
    }
    if ((newtick != apptTickMark) || (apptTickMark >= 0)) {
        WinEraseLine(0, r.topLeft.y, 0, r.topLeft.y+r.extent.y-1);
        if (newtick >= 0) {
            i = TblGetRowHeight(tableP, 0);
            WinDrawLine(0, r.topLeft.y+(i*newtick)+(i/4), 0,
                r.topLeft.y+(i*(newtick+1))-(i/4)-1);
        }
        apptTickMark = newtick;
    }
    MemPtrUnlock(li);
}

static Boolean MainFormHandleEvent(EventPtr e)
{
    static short whichside=0;
    static long tickticks=0;
    static Boolean forminit=false;
    static Boolean stillonmainform=true;
    long ticks;
    Boolean handled;
    FormPtr frm;
    ControlPtr ptr;
    DateType BeginDate={0, 0}, tempd;
    DWord ref;
    ULong alarmTime;
    UInt DOW;
    UInt cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    GoToParamsPtr theGotoPointer;
    Word matchFieldNum;
    short i1, day;
    UInt apptnum, rownum;
    UInt recordNum;
    LineItemPtr li;
    TablePtr tableP;
    ListPtr lst;
    ULong AppID, DataID;
    SWord selmonth, selday, selyear;
    VoidHand strH;
    CharPtr strP;
    RectangleType r1, r2, r3, r4; /* FME - May 6, 2000 Added r4 for prev/next */
    Word width;
    SWord x, y;
    Boolean ishighlit, isinside, penDown;
    short i;
    Int newSelection;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();

        /*
         * Calculate the width of the "Today" pushbutton and make it fill
         * the remaining space.  I do this here rather than statically in
         * the resources to allow for different languages.
         */

        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_tomorrow), &r2);
        r2.extent.x += 2;
        width = r2.extent.x+1;

        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_week), &r3);
        r3.extent.x += 2;
        width += r3.extent.x+1;

        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_wkprev), &r4);
        r4.extent.x += 2;
        width += r4.extent.x+1;

        FrmGetFormBounds(frm, &r1);
        width = r1.extent.x-width-2;

        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_today), &r1);

        r1.extent.x = width;
        FrmSetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_today), &r1);

        r2.topLeft.x = r1.topLeft.x+r1.extent.x+1;
        FrmSetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_tomorrow), &r2);

        r3.topLeft.x = r2.topLeft.x+r2.extent.x+1;
        FrmSetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_week), &r3);

        /* FME - 6 May, 2000: Adjust location for previous week pushbutton */
        r4.topLeft.x = r3.topLeft.x+r3.extent.x+1;
        FrmSetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_wkprev), &r4);

        /* FME - 6 May, 2000: Adjust location for next week pushbutton */
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_wknext), &r4);
        r4.extent.x += 2;
        r4.topLeft.x = r3.topLeft.x+r3.extent.x+1;
        FrmSetObjectBounds(frm, FrmGetObjectIndex(frm, pbuttonID_wknext), &r4);

        /*
         * Save off the original labels for the today & week tabs.  When these
         * are pressed I change them to the MM/DD (or equiv).
         */
        ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, pbuttonID_today));
        origtodaylabel = CtlGetLabel(ptr);
        ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, pbuttonID_week));
        origweeklabel = CtlGetLabel(ptr);
        if (lasttab == -1) {
            refreshlist((short) (PrefsR->ApptPrefs.defaulttab-'0'));
        } else {
            refreshlist(lasttab);
        }
        FrmDrawForm(frm);
        apptTickMark = -1;
        tickticks = TimGetTicks();
        DisplayFooter(true);
        forminit = true;
        handled = true;
        break;

    case frmCloseEvent:
        /* don't need these any more */
        freelistitems();
        forminit = false;
        break;

    case frmUpdateEvent:
        /*
         * Redraw the bottom section of the form if we get a form update
         * event.  Not very likely, but the OS3.5 debug ROM was built to
         * provide this often for testing your program's reaction to low
         * memory conditions.  We return handled = true so that the default
         * action of the OS doesn't call FrmDrawForm() again.
         */
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        DisplayFooter(true);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        case mitemID_apptprefs:
            FrmGotoForm(ApptPrefsForm);
            break;
        case mitemID_todoprefs:
            FrmGotoForm(ToDoPrefsForm);
            break;
        case mitemID_launchprefs:
            FrmGotoForm(LaunchPrefsForm);
            break;
        case mitemID_about:
            FrmPopupForm(AboutForm);
            break;
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case pbuttonID_today:
            frm = FrmGetActiveForm();
            /* Reset the week label */
            ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, pbuttonID_week));
            if (CtlGetLabel(ptr) != origweeklabel)
                CtlSetLabel(ptr, origweeklabel);
            if ((strH = DmGetResource(strRsc, strID_todaypicktitle)) == 0)
                break;
            strP = MemHandleLock(strH);
            /*
             * If the user taps the today tab while already on it, we allow
             * selection of a day.  We start off the pick list with the
             * currently selected day.
             */
            DateSecondsToDate(TimGetSeconds(), &BeginDate);
            tempd = BeginDate;
            if (lasttab == 0) {
                selmonth = apptDate.month;
                selday = apptDate.day;
                selyear = apptDate.year+1904;
                BeginDate = apptDate; /* for when cancel is hit */
            } else {
                selmonth = BeginDate.month;
                selday = BeginDate.day;
                selyear = BeginDate.year+1904;
            }
            if ((lasttab == 0) &&
                    SelectDay(selectDayByDay, &selmonth, &selday, &selyear,
                        strP)) {
                BeginDate.month = selmonth;
                BeginDate.day = selday;
                BeginDate.year = selyear-1904;
            }
            MemHandleUnlock(strH);
            DmReleaseResource(strH);
            /*
             * Create an appropriate date style label for the today tab.
             */
            DateToAscii(BeginDate.month, BeginDate.day, BeginDate.year+1904,
                Prefsdfmt2, AppErrStr);
            /* slash, dot, or dash */
            switch (Prefsdfmt2) {
            case dfYMDWithDots:
            case dfYMDWithDashes:
                i1 = StrLen(AppErrStr);
                StrNCopy(todaylabel, &AppErrStr[3], i1-3);
                StrNCopy(&todaylabel[i1-3], &AppErrStr[2], 1);
                StrNCopy(&todaylabel[i1-2], AppErrStr, 2);
                todaylabel[i1] = (char) NULL;
                break;
            case dfYMDWithSlashes:
            default:
                i1 = StrLen(AppErrStr);
                StrNCopy(todaylabel, AppErrStr, i1);
                todaylabel[i1] = (char) NULL;
                break;
            }
            if (lasttab == 0) {
                /*
                 * Only replace label if the day is changed.  This is different
                 * from the week label, which I alway set.  Kind of a personal
                 * preference thing.
                 */
                ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                    pbuttonID_today));
                /*
                 * If it is "Today" then reset the label.
                 */
                if (DateToInt(tempd) == DateToInt(BeginDate)) {
                    CtlSetLabel(ptr, origtodaylabel);
                } else {
                    CtlSetLabel(ptr, todaylabel);
                }
            }
            lasttab = 0;
            MainFormLoadTable(frm, 1, 0, BeginDate, 1);
            ShowScrollArrows(frm);
            /*
             * The tick mark is a little line that sits next to the closest
             * appointment.  It only does this for today.
             */
            ShowTickMark(frm, false);
            break;
        case pbuttonID_tomorrow:
            frm = FrmGetActiveForm();
            /* Reset the today label */
            ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, pbuttonID_today));
            if (CtlGetLabel(ptr) != origtodaylabel)
                CtlSetLabel(ptr, origtodaylabel);
            /* Reset the week label */
            ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, pbuttonID_week));
            if (CtlGetLabel(ptr) != origweeklabel)
                CtlSetLabel(ptr, origweeklabel);
            DateSecondsToDate(TimGetSeconds(), &BeginDate);
            /* warp to tomorrow */
            DateDaysToDate(DateToDays(BeginDate)+1, &BeginDate);
            lasttab = 1;
            MainFormLoadTable(frm, 1, 0, BeginDate, 1);
            ShowScrollArrows(frm);
            break;
        case pbuttonID_week:
            frm = FrmGetActiveForm();
            /* Reset the today label */
            ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, pbuttonID_today));
            if (CtlGetLabel(ptr) != origtodaylabel)
                CtlSetLabel(ptr, origtodaylabel);
            if ((strH = DmGetResource(strRsc, strID_weekpicktitle)) == 0)
                break;
            strP = MemHandleLock(strH);
            /*
             * If the user taps the week tab while already on it, we allow
             * selection of the week.  We start off the pick list with the
             * currently selected day -- if available (i.e. lasttab is 2).
             */
            DateSecondsToDate(TimGetSeconds(), &BeginDate);
            if (lasttab == 2) {
                selmonth = apptDate.month;
                selday = apptDate.day;
                selyear = apptDate.year+1904;
                BeginDate = apptDate; /* for when cancel is hit */
            } else {
                selmonth = BeginDate.month;
                selday = BeginDate.day;
                selyear = BeginDate.year+1904;
            }
            if ((lasttab == 2) &&
                    SelectDay(selectDayByDay, &selmonth, &selday, &selyear,
                        strP)) {
                BeginDate.month = selmonth;
                BeginDate.day = selday;
                BeginDate.year = selyear-1904;
            } else {
                if (PrefsR->ApptPrefs.weekstart == '1') {
                    /* find the previous Monday */
                    DOW = DayOfWeek(BeginDate.month, BeginDate.day,
                        BeginDate.year+1904);
                    if (DOW > 1) {
                        DateDaysToDate(DateToDays(BeginDate)-(DOW-1),
                            &BeginDate);
                    } else if (DOW == 0) {
                        /*
                         * If Sunday we'll go back to the previous Monday
                         * as well
                         */
                        DateDaysToDate(DateToDays(BeginDate)-6, &BeginDate);
                    }
                }
            }
            MemHandleUnlock(strH);
            DmReleaseResource(strH);

            /* Update week label */
            UpdateWeekLabel(frm, BeginDate);

            lasttab = 2;
            break;
        case buttonID_pageup:
            if (apptTop) {
                apptTop -= MIN(apptTop, apptnumRows);
                frm = FrmGetActiveForm();
                MainFormLoadTable(frm, 2, apptTop, apptDate, apptDays);
                ShowScrollArrows(frm);
                if (lasttab == 0) ShowTickMark(frm, false);
            }
            break;
        case buttonID_pagedown:
            if ((apptTop+apptnumRows) < apptTotal) {
                apptTop += apptnumRows;
                frm = FrmGetActiveForm();
                MainFormLoadTable(frm, 2, apptTop, apptDate, apptDays);
                ShowScrollArrows(frm);
                if (lasttab == 0) ShowTickMark(frm, false);
            }
            break;
        case pbuttonID_wkprev:
            frm = FrmGetActiveForm();

            /* Move backwards 7 days */
            DateDaysToDate(DateToDays(apptDate)-7, &BeginDate);

            /* Update week label */
            UpdateWeekLabel(frm, BeginDate);

            /*
             * FME - May 14, 2000: Fix -> Date selection dialog was wrongfully
             *                            launching pressing Today tab after
             *                            using +/- week
             *                     Fix -> Pressing the week tab after +/- week
             *                            jumped back week based on current
             *                            date, it now launches the date
             *                            selection dialog.
             */
            lasttab = 2;
            break;

        case pbuttonID_wknext:
            frm = FrmGetActiveForm();

            /* Move forward 7 days */
            DateDaysToDate(DateToDays(apptDate)+7, &BeginDate);

            /* Update week label */
            UpdateWeekLabel(frm, BeginDate);

            /*
             * FME - May 14, 2000: Fix -> Date selection dialog was wrongfully
             *                            launching pressing Today tab after
             *                            using +/- week
             *                     Fix -> Pressing the week tab after +/- week
             *                            jumped back week based on current
             *                            date, it now launches the date
             *                            selection dialog.
             */
            lasttab = 2;
            break;
        }
        break;

    case keyDownEvent:
        if (e->data.keyDown.chr == pageUpChr) {
            frm = FrmGetActiveForm();
            ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                buttonID_pageup));
            CtlHitControl(ptr);
        } else if (e->data.keyDown.chr == pageDownChr) {
            frm = FrmGetActiveForm();
            ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                buttonID_pagedown));
            CtlHitControl(ptr);
        }
        break;

   case tblEnterEvent:
        /*
         * We handle the ToDo title bar tap in here so it doesn't get a
         * chance to highlight the row.
         */
        frm = FrmGetActiveForm();
        tableP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            tableID_dates));
        rownum = TblGetRowID(tableP, e->data.tblSelect.row);
        if (rownum == apptAlarmLine) {
            SndPlaySystemSound(sndClick);
            if (apptTop != apptAlarmLine) {
                /* bring these to top if the untimed alarms title is tapped */
                apptTop = apptAlarmLine;
                MainFormLoadTable(frm, 2, apptTop, apptDate, apptDays);
                ShowScrollArrows(frm);
            }
            handled = true;
        } else if (rownum == apptToDoLine) {
            SndPlaySystemSound(sndClick);
            /*
             * A click on the ToDo title within the table lets you change
             * the category being displayed.  This is only available during
             * the single day views because I don't show ToDos on the week
             * view.
             *
             * But just clicking on the left side of the title will bring
             * it to the top.
             */
            if (whichside) {
                SelectCategoryPopup(ToDoDB, tableP, e->data.tblSelect.row);
            } else if ((PrefsR->ToDoPrefs.gotop == '1') &&
                (apptTop != apptToDoLine)) {
                apptTop = apptToDoLine;
                MainFormLoadTable(frm, 2, apptTop, apptDate, apptDays);
                ShowScrollArrows(frm);
            } else {
                /*
                 * Clicking on the left side while the title is already
                 * at the top will cycle to the next category.
                 */
                apptToDoCategory = CategoryGetNext(ToDoDB,
                    apptToDoCategory);
                /*
                 * Copy the category name into my preferences record and
                 * mark it dirty.
                 */
                MemSet(PrefsR->todocategory, sizeof(PrefsR->todocategory),
                    0);
                CategoryGetName(ToDoDB, apptToDoCategory,
                    PrefsR->todocategory);
                /*
                 * No immediate need to write, so we just flag it for when
                 * the program ends.
                 */
                PrefsRdirty = true;
                if (PrefsR->ToDoPrefs.gotop == '1') apptTop = apptToDoLine;
                /*
                 * This is a quickie regen of the table data.  It only
                 * does the ToDo items because that is the only thing
                 * that has changed and we want it to work quickly.
                 */
                MainFormLoadTable(frm, 3, apptTop, apptDate, apptDays);
                ShowScrollArrows(frm);
            }
            handled = true;
        }
        break;

    case tblSelectEvent:
        frm = FrmGetActiveForm();
        tableP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, tableID_dates));
        rownum = TblGetRowID(tableP, e->data.tblSelect.row);
        apptnum = rownum;
        for (day=0; day < apptDays; day++) {
            if (apptnum < apptCounts[day]) break;
            apptnum -= apptCounts[day];
        }
        if ((day >= apptDays) ||
            !(li = MemHandleLock(apptLists[day]))) {
            break;
        }
        /*
         * This is a copy of the LineItemType entry.  Mostly I use this
         * for the Notes popup.  I load it into a ApptTypeInfo because I
         * want to have the .type indicator.
         */
        apptCurrent.recordNum = li[apptnum].recordNum;
        apptCurrent.date = li[apptnum].date;
        recordNum = li[apptnum].recordNum;
        MemPtrUnlock(li);
        if (rownum > apptToDoLine) {
            /* todo */
            DataID = ToDoAppID;
            AppID = Prefstodolaunch;
            apptCurrent.type = 1;
        } else if (rownum < apptToDoLine) {
            /* appointment */
            DataID = DatebookAppID;
            AppID = Prefsapptlaunch;
            apptCurrent.type = 0;
        } else break;
        matchFieldNum = 0;
        if (whichside) {
            /*
             * Right side is for notes.  But first see if I indicated that
             * there was a note.
             */
            if (TblGetItemInt(tableP, e->data.tblSelect.row, 0))
                apptCurrentNote = true;
            else apptCurrentNote = false;

            /*
             * Tapped directly on the note icon.  So we'll launch the app
             * right into the note rather than showing our own popup.
             */
            if ((whichside == 2) && apptCurrentNote) {
                matchFieldNum = 1; /* note */
            } else {
                /* I use this for popup positioning */
                apptCurrentRow = e->data.tblSelect.row;
                FrmPopupForm(NotesForm);
                break;
            }
        }
        /* only the left side of the screen is for application launch */
        /* Allocate the new pointer for the goto parameters. */
        theGotoPointer = MemPtrNew(sizeof(GoToParamsType));
        if (!theGotoPointer) break;
        /*
         * Set the owner of the pointer to be the system. This is
         * required because all memory allocated by our application
         * is freed when the application quits. Our application will
         * quit the next time through our event handler.
         *
         * Grab the id of the database we want the record to come from.  This
         * is always going to be DatebookDB or ToDoDB.
         */
        if ((MemPtrSetOwner(theGotoPointer, 0) == 0) &&
            (DmGetNextDatabaseByTypeCreator(true, &searchInfo,
            'DATA', DataID, true, &cardNo, &dbID) == 0)) {
            /*
             * Copy all the goto information into the GoToParamsPtr
             * structure.
             */
            theGotoPointer->searchStrLen = 0;
            theGotoPointer->dbCardNo = cardNo;
            theGotoPointer->dbID = dbID;
            theGotoPointer->recordNum = recordNum;
            theGotoPointer->matchPos = 0;
            theGotoPointer->matchFieldNum = matchFieldNum;
            theGotoPointer->matchCustom = 0;
            /*
             * Grab the id of the application we want to launch.  This is
             * defined by preferences.  Who knows what'll happen if the
             * launched program doesn't like the structure.  ToDo seems
             * to be especially unstable if launched with a Datebook
             * record.  That's because they don't verify if the record
             * number is correct, and they assume it is always within
             * ToDoDB!  I'll assign the punishment -- castration!
             *
             * Hrm, looking at both Datebook and ToDo source it appears that
             * there is just no error testing at all to control a condition
             * where the record number passed is out of range.  So what am
             * I supposed to do!?!  I don't want the device to fatal error
             * just because the user selected an inappropriate application
             * to launch.
             *
             * The only thing I can do is limit the list to known working
             * applications!  God what a horrible situation.
             *
             * Okay, here's the compromise..  I'll only GoTo on certain
             * applications, and everthing else I will NormalLaunch.
             */
            if (((DataID == DatebookAppID) &&
                ((AppID == 'date') ||                   /* built-in Datebook */
                (AppID == 'Actn') ||                    /* ActionNames */
                (AppID == 'CESD') ||                    /* DateBk3 */
                (AppID == 'HsDB'))) ||                  /* DateBk3 Handspring */
                ((DataID == ToDoAppID) &&
                ((AppID == 'todo') ||                   /* built-in ToDo */
                (AppID == 'G200')))) {                  /* ToDo PLUS */
                if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
                    sysFileTApplication, AppID, true, &cardNo, &dbID)
                    == 0)) {
                    SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdGoTo,
                    (Ptr) theGotoPointer);
                }
            } else if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
                sysFileTApplication, AppID, true, &cardNo, &dbID) == 0)) {
                SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdNormalLaunch, NULL);
            }
        } else {
            MemPtrFree(theGotoPointer);
        }
        break;

    case penDownEvent:
        /*
         * This is when we want to ignore the next tap on this form.
         */
        if (ignorependown) {
            ignorependown = false;
            handled = true;
            break;
        }
        /*
         * This is a pretty cheezy way to detect a tap in the lower
         * left corner.  I would've put a gadget down there, but then
         * I would have probably wanted to draw the date/time using
         * special gadget routines.
         */
        if ((e->screenX < 94) && (e->screenY > 137)) {
            SndPlaySystemSound(sndClick);
            if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
                sysFileTApplication, Prefslowerleftlaunch, true, &cardNo,
                &dbID) == 0)) {
                SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdNormalLaunch, NULL);
            }
        }
        /*
         * Couldn't figure out how to get the tap position within a table
         * select.  Since I want to distinguish between tapping the left
         * side or the right side, I will store which side was last tapped
         * into a static variable.
         */
        FntSetFont(symbolFont);
        i1 = FntCharWidth('\013');
        FntSetFont(stdFont);
        /* See if the user has tapped in the vacinity of a note icon. */
        if (e->screenX >= (160-i1)) whichside = 2;
        else if (e->screenX >= 80) whichside = 1;
        else whichside = 0;

        /*
         * Handle the clock and phone icons.
         */
        frm = FrmGetActiveForm();
        /*
         * Checking if the user is pressing one of the icons.  I
         * have gadgets to represent where they should be.
         */
        isinside = true;
        do {
        i = 0;
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_clock), &r1);
        if (RctPtInRectangle(e->screenX, e->screenY, &r1)) break;
        i++;
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_phone), &r1);
        if (RctPtInRectangle(e->screenX, e->screenY, &r1)) break;
        isinside = false;
        } while (0);
        if (!isinside) break;
        ishighlit = false;
        /*
         * While the pen is held down I stay in this loop.  The user can move
         * the stylus outside of the area to stop the operation.
         */
        do {
            PenGetPoint(&x, &y, &penDown);
            isinside = RctPtInRectangle(x, y, &r1);
            if ((!ishighlit && isinside) ||
                (ishighlit && !isinside)) {
                WinInvertRectangle(&r1, 3);
                ishighlit = !ishighlit;
            }
        } while (penDown);
        /*
         * If the stylus was still inside the icon when released then launch
         * the application.
         */
        if (ishighlit) {
            WinInvertRectangle(&r1, 3);
            SndPlaySystemSound(sndClick);
            switch (i) {
            case 0: /* clock */
                lst = FrmGetObjectPtr(frm,
                    FrmGetObjectIndex(frm, listID_times));
                newSelection = LstPopupList(lst);
                if ((newSelection >= 0) && (newSelection < TIMESINLIST)) {
                    alarmTime = TimGetSeconds()+times[newSelection];
                    /* 0 means off, but make sure it was previously set */
                    if (newSelection == 0) {
                        if (GetTimeOfNextAlarm(&ref) == 0) break;
                        alarmTime = 0;
                    }
                    SetTimeOfNextAlarm(alarmTime, 0);
                }
                break;
            case 1: /* phone (contacts) */
                newSelection = ContactsPopup(listID_contacts);
                break;
            }
        }
        break;

    /*
     * This is used to recognize when the menu is active.  Kind of an after
     * thought because I didn't realize that the form handler function still
     * received events while the menu was active.
     */
    case winEnterEvent:
        if (e->data.winEnter.enterWindow ==
            (WinHandle) FrmGetFormPtr(MainForm))
            stillonmainform = true; else stillonmainform = false;
        break;

    case nilEvent:
        /* try not to do this until frmOpenEvent happens */
        if (!forminit || !stillonmainform) break;
        /*
         * Flash the appointment tick mark every second.  And update
         * the footer as well.
         */
        ticks = TimGetTicks();
        if ((ticks-tickticks) >= (tps*1)) {
            tickticks = ticks;
            if (lasttab == 0) {
                frm = FrmGetActiveForm();
                if (apptTickMark == -1) {
                    ShowTickMark(frm, false);
                } else {
                    ShowTickMark(frm, true);
                }
            }
            DisplayFooter(false);
            DateSecondsToDate(TimGetSeconds(), &BeginDate);
            if (DateToInt(BeginDate) != DateToInt(CurrentDate)) {
                /*
                 * If we see that the day has rolled while we are on
                 * we'll refresh the list.  This is going to be rare
                 * because you would have to be staring at the screen
                 * at midnight to notice.
                 */
                DateSecondsToDate(TimGetSeconds(), &CurrentDate);
                refreshlist(-1);
            }
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

/*
 * Make sure str is large enough to hold the string.  It will look like:
 *
 * Sat Jun 12, 1999 10:00p - 11:00p
 */
static void DateToString(char *str, DateType d, TimeType t1, TimeType t2)
{
    short i1, i2;
    if (DateToInt(d) == toDoNoDueDate) {
        StrCopy(str, "No Date");
    } else {
        DateToDOWDMFormat(d.month, d.day, d.year+1904, Prefsdfmt1, str);
    }
    if (TimeToInt(t1) != apptNoTime) {
        StrCat(str, " ");
        i1 = StrLen(str);
        TimeToAscii(t1.hours, t1.minutes, Prefstfmt, &str[i1]);
        i2 = StrLen(str);
        if ((i2-i1) >= 4) {
            /* move the a or p in a little closer, if it exists */
            if ((UCASE(str[i2-2]) == 'A') ||
                (UCASE(str[i2-2]) == 'P')) {
                str[i2-3] = str[i2-2];
                i2 -= 2;
                str[i2] = (char) NULL;
            }
        }
        if (TimeToInt(t2) != apptNoTime) {
            StrCat(str, " - ");
            i1 = StrLen(str);
            TimeToAscii(t2.hours, t2.minutes, Prefstfmt, &str[i1]);
            i2 = StrLen(str);
            if ((i2-i1) >= 4) {
                /* move the a or p in a little closer, if it exists */
                if ((UCASE(str[i2-2]) == 'A') ||
                    (UCASE(str[i2-2]) == 'P')) {
                    str[i2-3] = str[i2-2];
                    i2 -= 2;
                    str[i2] = (char) NULL;
                }
            }
        }
    }
}

static Boolean NotesFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    FieldPtr fld;
    ApptPackedDBRecordPtr apptrecP;
    VoidPtr rp;
    VoidHand recHandle;
    long l1, l2, l3, l4;
    Int width, width2, textfit;
    short titlerows, titlerows2;
    UInt y;
    Boolean fits;
    RectangleType r, r2, r3;
    TimeType nadatime={-1, -1};
    static SWord penY, deltaY, h;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        WinGetWindowFrameRect((WinHandle) frm, &r3);
        /*
         * Get the form rectangle, but only consider the width to be valid
         * at this time because I may resize this below.
         */
        FrmGetFormBounds(frm, &r);
        /*
         * If we are running in POSER make sure to set the notes
         * window is full width to avoid a debug ROM warning.
         */
        if (!inposer) {
            r3.topLeft.x += 4;
            r3.extent.x -= 8;
            r.topLeft.x += 4;
            r.extent.x -= 8;
        }
        r.topLeft.x = 1;
        r.extent.x -= 3;
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_notesnote));
        switch (apptCurrent.type) {
        case 0:
            /* appointment */
            if ((recHandle = DmQueryRecord(DatebookDB,
                apptCurrent.recordNum))) {
                apptrecP = MemHandleLock(recHandle);
                /* Calculate the position of the description */
                l1 = sizeof(ApptDateTimeType)+sizeof(ApptDBRecordFlags);
                if (apptrecP->flags.alarm) l1 += sizeof(AlarmInfoType);
                if (apptrecP->flags.repeat) l1 += sizeof(RepeatInfoType);
                if (apptrecP->flags.exceptions)
                    l1 += sizeof(UInt) +
                        (((ExceptionsListType *)
                        ((CharPtr) apptrecP+l1))->numExceptions*
                        sizeof(DateType));
                /*
                 * The reason this code is quite ugly is because I need to
                 * do some dynamic sizing of the window depending on whether
                 * there is a note and how many lines the title will be.
                 */
                titlerows = 0;
                FntSetFont(boldFont);
                if (apptrecP->flags.description) {
                    l2 = StrLen((CharPtr) apptrecP+l1);
                    l3 = 0;
                    /* how many lines for title */
                    fits = false;
                    while (l2 && (!fits) && (titlerows < 5)) {
                        /* newlines don't work with FntCharsinWidth() */
                        if (*((CharPtr) apptrecP+l1+l3) == '\n') {
                            textfit = 1;
                            fits = false;
                        } else {
                            textfit = l2;
                            width = r.extent.x;
                            FntCharsInWidth((CharPtr) apptrecP+l1+l3, &width,
                                &textfit, &fits);
                            /*
                             * Sometimes it doesn't like spaces and it will
                             * tell us it doesn't fit.  So we skip it.
                             */
                            if (width == 0) textfit = 1; else titlerows++;
                        }
                        l3 += textfit;
                        l2 -= textfit;
                    }
                }
                /*
                 * If there isn't a note, shrink the window down.
                 */
                if (!apptCurrentNote) {
                    r3.extent.y = 4+(titlerows*FntCharHeight());
                    /* room for the date/time line */
                    FntSetFont(stdFont);
                    r3.extent.y += FntCharHeight()+2;
                    /*
                     * Try and place the small popup underneath the
                     * highlight row.
                     */
                    if (apptCurrentRow < 6) {
                        r3.topLeft.y = 16+(FntCharHeight()*(apptCurrentRow+1))+
                            (FntCharHeight()/2);
                    } else {
                        r3.topLeft.y = 16+(FntCharHeight()*apptCurrentRow)-
                            r3.extent.y-(FntCharHeight()/2);
                    }
                    FntSetFont(boldFont);
                }
                /*
                 * If we are running in POSER make sure to set the notes
                 * window flush with the bottom to avoid a debug ROM warning.
                 */
                if (inposer) {
                    r3.topLeft.y = 160-r3.extent.y;
                }
                WinSetWindowBounds((WinHandle) frm, &r3);
                FrmDrawForm(frm);
                FrmGetFormBounds(frm, &r);
                r.topLeft.x = 1;
                r.topLeft.y = 1;
                r.extent.x -= 3;
                r.extent.y -= 3;
                WinDrawRectangleFrame(popupFrame, &r);
                r2 = r;
                r2.extent.y = (titlerows*FntCharHeight())+1;
                WinDrawRectangle(&r2, 0);
                y = 1;
                if (apptrecP->flags.description) {
                    l2 = StrLen((CharPtr) apptrecP+l1);
                    l3 = 0;
                    l4 = l2;
                    fits = false;
                    titlerows2 = titlerows;
                    while (l4 && (!fits) && titlerows2) {
                        /* newlines don't work with FntCharsinWidth() */
                        if (*((CharPtr) apptrecP+l1+l3) == '\n') {
                            textfit = 1;
                            fits = false;
                        } else {
                            textfit = l4;
                            width = r.extent.x;
                            FntCharsInWidth((CharPtr) apptrecP+l1+l3, &width,
                                &textfit, &fits);
                            /* center if one line, otherwise left aligned */
                            if (titlerows > 1) {
                                WinDrawInvertedChars((CharPtr) apptrecP+l1+l3,
                                    textfit, 2, y);
                            } else {
                                WinDrawInvertedChars((CharPtr) apptrecP+l1+l3,
                                    textfit, (r.extent.x/2)-(width/2)+2, y);
                            }
                            /*
                             * Sometimes it doesn't like spaces and it will
                             * tell us it doesn't fit.  So we skip it.
                             */
                            if (width == 0) textfit = 1;
                            else {
                                y += FntCharHeight();
                                titlerows2--;
                            }
                        }
                        l3 += textfit;
                        l4 -= textfit;
                    }
                    l1 += l2+1;
                    y++;
                }

                /* Put a little alarm icon on the right */
                if (apptrecP->flags.alarm) {
                    FntSetFont(symbolFont);
                    width2 = FntCharWidth('\024');
                    WinDrawChars("\024", 1, r.extent.x-width2, y+1);
                } else width2 = 0;

                /* The date & times */
                FntSetFont(stdFont);
                DateToString(AppErrStr, apptCurrent.date,
                    apptrecP->when.startTime, apptrecP->when.endTime);
                l2 = StrLen(AppErrStr);
                width = r.extent.x-width2;
                textfit = l2;
                FntCharsInWidth(AppErrStr, &width, &textfit, &fits);
                WinDrawChars(AppErrStr, textfit, ((r.extent.x-width2)/2)-
                    (width/2)+2, y+1);

                y += FntCharHeight()+2;
                if (apptCurrentNote && apptrecP->flags.note) {
                    WinDrawLine(4, y, r.extent.x-3, y);
                    y += 2;
                    /* position the notes field */
                    FldGetBounds(fld, &r3);
                    r3.topLeft.y = y;
                    r3.extent.y = r.extent.y-r3.topLeft.y;
                    FldSetBounds(fld, &r3);
                    FldSetText(fld, recHandle, l1,
                        StrLen((CharPtr) apptrecP+l1)+1);
                    FldDrawField(fld);
                }
                /* Unlock the handle to the record */
                MemHandleUnlock(recHandle);
            }
            break;
        case 1:
            /* todo */
            if ((recHandle = DmQueryRecord(ToDoDB, apptCurrent.recordNum))) {
                rp = MemHandleLock(recHandle);
                /* Calculate the position of the description */
                l1 = sizeof(DateType)+sizeof(unsigned char);
                /*
                 * The reason this code is quite ugly is because I need to
                 * do some dynamic sizing of the window depending on whether
                 * there is a note and how many lines the title will be.
                 */
                titlerows = 0;
                FntSetFont(boldFont);
                l2 = StrLen((CharPtr) rp+l1);
                if (l2) {
                    l3 = 0;
                    /* how many lines for title */
                    fits = false;
                    while ((!fits) && (titlerows < 5)) {
                        /* newlines don't work with FntCharsinWidth() */
                        if (*((CharPtr) rp+l1+l3) == '\n') {
                            textfit = 1;
                        } else {
                            textfit = l2;
                            width = r.extent.x;
                            FntCharsInWidth((CharPtr) rp+l1+l3, &width,
                                &textfit, &fits);
                            /*
                             * Sometimes it doesn't like spaces and it will
                             * tell us it doesn't fit.  So we skip it.
                             */
                            if (width == 0) textfit = 1; else titlerows++;
                        }
                        l3 += textfit;
                        l2 -= textfit;
                    }
                }
                /*
                 * If there isn't a note, shrink the window down.
                 */
                if (!apptCurrentNote) {
                    r3.extent.y = 4+(titlerows*FntCharHeight());
                    /* room for the date/time line */
                    FntSetFont(stdFont);
                    r3.extent.y += FntCharHeight()+2;
                    /*
                     * Try and place the small popup underneath the
                     * highlight row.
                     */
                    if (apptCurrentRow < 6) {
                        r3.topLeft.y = 16+(FntCharHeight()*(apptCurrentRow+1))+
                            (FntCharHeight()/2);
                    } else {
                        r3.topLeft.y = 16+(FntCharHeight()*apptCurrentRow)-
                            r3.extent.y-(FntCharHeight()/2);
                    }
                    FntSetFont(boldFont);
                }
                /*
                 * If we are running in POSER make sure to set the notes
                 * window flush with the bottom to avoid a debug ROM warning.
                 */
                if (inposer) {
                    r3.topLeft.y = 160-r3.extent.y;
                }
                WinSetWindowBounds((WinHandle) frm, &r3);
                FrmDrawForm(frm);
                FrmGetFormBounds(frm, &r);
                r.topLeft.x = 1;
                r.topLeft.y = 1;
                r.extent.x -= 3;
                r.extent.y -= 3;
                WinDrawRectangleFrame(popupFrame, &r);
                r2 = r;
                r2.extent.y = (titlerows*FntCharHeight())+1;
                WinDrawRectangle(&r2, 0);
                y = 1;
                l2 = StrLen((CharPtr) rp+l1);
                if (l2) {
                    l3 = 0;
                    l4 = l2;
                    fits = false;
                    titlerows2 = titlerows;
                    while ((!fits) && titlerows2) {
                        /* newlines don't work with FntCharsinWidth() */
                        if (*((CharPtr) rp+l1+l3) == '\n') {
                            textfit = 1;
                        } else {
                            textfit = l4;
                            width = r.extent.x;
                            FntCharsInWidth((CharPtr) rp+l1+l3, &width,
                                &textfit, &fits);
                            /* center if one line, otherwise left aligned */
                            if (titlerows > 1) {
                                WinDrawInvertedChars((CharPtr) rp+l1+l3,
                                    textfit, 2, y);
                            } else {
                                WinDrawInvertedChars((CharPtr) rp+l1+l3,
                                    textfit, (r.extent.x/2)-(width/2)+2, y);
                            }
                            /*
                             * Sometimes it doesn't like spaces and it will
                             * tell us it doesn't fit.  So we skip it.
                             */
                            if (width == 0) textfit = 1;
                            else {
                                y += FntCharHeight();
                                titlerows2--;
                            }
                        }
                        l3 += textfit;
                        l4 -= textfit;
                    }
                    l1 += l2+1;
                    y++;
                }
                /* The date & times */
                FntSetFont(stdFont);
                DateToString(AppErrStr, apptCurrent.date, nadatime, nadatime);
                l2 = StrLen(AppErrStr);
                width = r.extent.x;
                textfit = l2;
                FntCharsInWidth(AppErrStr, &width, &textfit, &fits);
                WinDrawChars(AppErrStr, textfit, (r.extent.x/2)-(width/2)+2,
                    y+1);
                y += FntCharHeight()+2;
                if (apptCurrentNote && ((MemHandleSize(recHandle)-l1) > 1)) {
                    /* there must also be a note on this record */
                    WinDrawLine(4, y, r.extent.x-3, y);
                    y += 2;
                    /* position the notes field */
                    FldGetBounds(fld, &r3);
                    r3.topLeft.y = y;
                    r3.extent.y = r.extent.y-r3.topLeft.y;
                    FldSetBounds(fld, &r3);
                    FldSetText(fld, recHandle, l1,
                        StrLen((CharPtr) rp+l1)+1);
                    FldDrawField(fld);
                }
                /* Unlock the handle to the record */
                MemHandleUnlock(recHandle);
            }
            break;
        }
        FntSetFont(stdFont);
        penY = 0;
        handled = true;
        break;

    case frmCloseEvent:
        frm = FrmGetActiveForm();
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_notesnote));
        FldSetTextHandle(fld, 0);
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case fieldID_notesnote:
            /*
             * This keeps the text from selecting so that my finger
             * scrolling will work.
             */
            handled = true;
        }
        break;

    case keyDownEvent:
        if (e->data.keyDown.chr == pageUpChr) {
            frm = FrmGetActiveForm();
            fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_notesnote));
            if (FldScrollable(fld, up))
                FldScrollField(fld, FldGetVisibleLines(fld), up);
        } else if (e->data.keyDown.chr == pageDownChr) {
            frm = FrmGetActiveForm();
            fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_notesnote));
            if (FldScrollable(fld, down))
                FldScrollField(fld, FldGetVisibleLines(fld), down);
        }
        break;

    case penDownEvent:
        penY = e->screenY;
        handled = true;
        break;

    case penMoveEvent:
        frm = FrmGetActiveForm();
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            fieldID_notesnote));
        deltaY = e->screenY-penY;
        h = FntCharHeight()/2;
        if (deltaY < -h) {
            if (FldScrollable(fld, down))
                FldScrollField(fld, -(deltaY/h), down);
            penY = e->screenY;
        } else if (deltaY > h) {
            if (FldScrollable(fld, up)) FldScrollField(fld, deltaY/h, up);
            penY = e->screenY;
        }
        break;

    case winExitEvent:
        /*
         * I hate scrollbars!  So I am just letting the user drag his finger
         * up and down to scroll the note.  If the screen is tapped outside
         * of the popup we will get rid of the notes popup.
         */
        frm = FrmGetActiveForm();
        if ((FormPtr) e->data.winExit.exitWindow == frm) {
            /*
             * Since FrmReturnToForm() doesn't fire a frmCloseEvent I have to
             * do this here as well.
             */
            frm = FrmGetActiveForm();
            fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_notesnote));
            FldSetTextHandle(fld, 0);
            ignorependown = true;
            FrmReturnToForm(MainForm);
            handled = true;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean ApptPrefsFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    static Boolean save;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        LoadApptPrefsFields();
        FrmDrawForm(frm);
        save = false;
        handled = true;
        break;

    case frmCloseEvent:
        UnloadApptPrefsFields(save);
        if (save) WritePrefsRec();
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_apptprefsok:
            save = true;
            FrmGotoForm(MainForm);
            break;
        case buttonID_apptprefscancel:
            FrmGotoForm(MainForm);
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean ToDoPrefsFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    static Boolean save;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        LoadToDoPrefsFields();
        FrmDrawForm(frm);
        save = false;
        handled = true;
        break;

    case frmCloseEvent:
        UnloadToDoPrefsFields(save);
        if (save) WritePrefsRec();
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_todoprefsok:
            save = true;
            FrmGotoForm(MainForm);
            break;
        case buttonID_todoprefscancel:
            FrmGotoForm(MainForm);
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean LaunchPrefsFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    ULong AppID, oldAppID;
    static Boolean save;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        LoadLaunchPrefsFields();
        FrmDrawForm(frm);
        save = false;
        handled = true;
        break;

    case frmCloseEvent:
        UnloadLaunchPrefsFields(save);
        if (save) WritePrefsRec();
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case popuptriggerID_launchprefsapptlaunch:
            MemMove(&AppID, PrefsR->LaunchPrefs.apptlaunch, sizeof(AppID));
            Prefsapptlaunch = AppsPopup(listID_launchprefsapptlaunch,
                popuptriggerID_launchprefsapptlaunch, AppID);
            break;
        case popuptriggerID_launchprefstodolaunch:
            MemMove(&AppID, PrefsR->LaunchPrefs.todolaunch, sizeof(AppID));
            oldAppID = AppID;
            Prefstodolaunch = AppsPopup(listID_launchprefstodolaunch,
                popuptriggerID_launchprefstodolaunch, AppID);
            break;
        case popuptriggerID_launchprefslowerleftlaunch:
            MemMove(&AppID, PrefsR->LaunchPrefs.lowerleftlaunch, sizeof(AppID));
            oldAppID = AppID;
            Prefslowerleftlaunch = AppsPopup(listID_launchprefslowerleftlaunch,
                popuptriggerID_launchprefslowerleftlaunch, AppID);
            break;
        case popuptriggerID_launchprefscontactslaunch:
            MemMove(&AppID, PrefsR->LaunchPrefs.contactslaunch, sizeof(AppID));
            oldAppID = AppID;
            Prefscontactslaunch = AppsPopup(listID_launchprefscontactslaunch,
                popuptriggerID_launchprefscontactslaunch, AppID);
            break;
        case popuptriggerID_launchprefshard1launch:
            MemMove(&AppID, PrefsR->LaunchPrefs.hardlaunch[0], sizeof(AppID));
            oldAppID = AppID;
            Prefshardlaunch[0] = AppsPopup(listID_launchprefshard1launch,
                popuptriggerID_launchprefshard1launch, AppID);
            break;
        case popuptriggerID_launchprefshard2launch:
            MemMove(&AppID, PrefsR->LaunchPrefs.hardlaunch[1], sizeof(AppID));
            oldAppID = AppID;
            Prefshardlaunch[1] = AppsPopup(listID_launchprefshard2launch,
                popuptriggerID_launchprefshard2launch, AppID);
            break;
        case popuptriggerID_launchprefshard3launch:
            MemMove(&AppID, PrefsR->LaunchPrefs.hardlaunch[2], sizeof(AppID));
            oldAppID = AppID;
            Prefshardlaunch[2] = AppsPopup(listID_launchprefshard3launch,
                popuptriggerID_launchprefshard3launch, AppID);
            break;
        case popuptriggerID_launchprefshard4launch:
            MemMove(&AppID, PrefsR->LaunchPrefs.hardlaunch[3], sizeof(AppID));
            oldAppID = AppID;
            Prefshardlaunch[3] = AppsPopup(listID_launchprefshard4launch,
                popuptriggerID_launchprefshard4launch, AppID);
            break;
        case buttonID_launchprefsok:
            save = true;
            FrmGotoForm(MainForm);
            break;
        case buttonID_launchprefscancel:
            FrmGotoForm(MainForm);
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean AboutFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_aboutok:
            FrmReturnToForm(MainForm);
            handled = true;
            break;
        case buttonID_aboutcontrib:
            FrmPopupForm(ContribForm);
            handled = true;
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean ContribFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    ListPtr lst;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID_contrib));
        /* turn off the highlight */
        LstSetSelection(lst, -1);
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_contribok:
            FrmReturnToForm(AboutForm);
            handled = true;
            break;
        }
        break;

    case keyDownEvent:
        if (e->data.keyDown.chr == pageUpChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_contrib));
            LstScrollList(lst, up, LstGetVisibleItems(lst));
        } else if (e->data.keyDown.chr == pageDownChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_contrib));
            LstScrollList(lst, down, LstGetVisibleItems(lst));
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
        formId = e->data.frmLoad.formID;
        frm = FrmInitForm(formId);
        FrmSetActiveForm(frm);

        switch (formId) {
        case MainForm:
            FrmSetEventHandler(frm, MainFormHandleEvent);
            break;
        case NotesForm:
            FrmSetEventHandler(frm, NotesFormHandleEvent);
            break;
        case ApptPrefsForm:
            FrmSetEventHandler(frm, ApptPrefsFormHandleEvent);
            break;
        case ToDoPrefsForm:
            FrmSetEventHandler(frm, ToDoPrefsFormHandleEvent);
            break;
        case LaunchPrefsForm:
            FrmSetEventHandler(frm, LaunchPrefsFormHandleEvent);
            break;
        case AboutForm:
            FrmSetEventHandler(frm, AboutFormHandleEvent);
            break;
        case ContribForm:
            FrmSetEventHandler(frm, ContribFormHandleEvent);
            break;
        case EditContactsForm:
            FrmSetEventHandler(frm, EditContactsHandleEvent);
            break;
        }
        handled = true;
    }

    return handled;
}

static void freelistitems(void)
{
    short i;
    for (i=0; i < MAXDAYS; i++) {
        if (apptLists[i]) {
            MemHandleFree(apptLists[i]);
            apptLists[i] = 0;
            apptCounts[i] = 0;
        }
    }
}

/*
 * These are called in StartApplication and StopApplication because they
 * are required memory allocations that exist from start to finish of the
 * application run.
 */
static void freememories(void)
{
    freelistitems();
    if (PrefsRHandle) {
        MemHandleFree(PrefsRHandle);
        PrefsRHandle = 0;
    }
}

/* Get preferences, open (or create) app database */
static Word StartApplication(void)
{
    short i;
    DWord temp;

    PrefsRHandle = 0;
    tps = SysTicksPerSecond();
    for (i=0; i < MAXDAYS; i++) {
        apptLists[i] = 0;
        apptCounts[i] = 0;
    }
    if ((PrefsRHandle = MemHandleNew(sizeof(struct sPrefsR))) == 0) {
        freememories();
        return 1;
    }
    /*
     * Lock the PrefsR structure
     */
    PrefsR = MemHandleLock(PrefsRHandle);

    /*
     * Check if we are running inside of POSER.  I do this because I may
     * want to fix a few little quirks during testing that the debug ROM
     * will complain about.. but which I find are really just optimizations.
     *
     * Such as the requirement for windows always being flush with the bottom
     * of the screen.  I like the notes window to float around, but in some
     * cases the debug ROM will throw a fatal warning but it really isn't
     * something that should be fatal etc etc.
     */

    if (FtrGet('pose', 0, &temp) == ftrErrNoSuchFeature) inposer = false;
    else inposer=true;

    /* Get the system preferences structure */
    PrefGetPreferences(&sysPrefs);

    if (!OpenDatabases()) {
        if (!DatebookDB) {
            /*
             * BTW make sure there is " " (single space) in unused ^1 ^2 ^3
             * entries or PalmOS <= 2 will blow up.
             */
            FrmCustomAlert(ErrorAlert, "Please run Date Book once first.",
                " ", " ");
        }
        freememories();
        return 1;
    }
    /*
     * Here's what the current date is when we started.  I use this to
     * determine if the device has been turned on and it is a different day.
     * This means I need to update the current view.
     */
    DateSecondsToDate(TimGetSeconds(), &CurrentDate);
    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void)
{
    FrmSaveAllForms();
    FrmCloseAllForms();
    CloseDatabases();
    freememories();
}

/*
 * Refresh the list.  Usually called when the day has rolled over.
 *
 * 5/22/00 cff
 * I changed this to only update the list if the user is on Today
 * (no picked date) or Tomorrow.
 */
static void refreshlist(short new_tab)
{
    FormPtr frm;
    ControlPtr ptr;
    
    frm = FrmGetActiveForm();
    if (new_tab == -1) {
        /* invalidate the cache */
        PrefsR->DayCache.listcount[0] = (char) 0;
        PrefsR->DayCache.listcount[1] = (char) 0;
    }
    if ((new_tab == 0) ||
        CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        pbuttonID_today)))) {
        if (new_tab >= 0) {
            CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                pbuttonID_today)), 1);
        }
        /* Only refresh if it is still on Today */
        ptr = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, pbuttonID_today));
        if (CtlGetLabel(ptr) == origtodaylabel) {
            CtlHitControl(FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, pbuttonID_today)));
        }
    } else if ((new_tab == 1) || CtlGetValue(FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, pbuttonID_tomorrow)))) {
        if (new_tab >= 0) {
            CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                pbuttonID_tomorrow)), 1);
        }
        CtlHitControl(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, pbuttonID_tomorrow)));
    } else if ((new_tab == 2) || CtlGetValue(FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, pbuttonID_week)))) {
        if (new_tab >= 0) {
            CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                pbuttonID_week)), 1);
        }
        CtlHitControl(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, pbuttonID_week)));
    }
    /*
     * I set this to -1 so that I don't get a double-bounce upon list
     * refresh.  By this I mean that the Week tab would bring up the
     * pick a day dialog if I left lasttab=2.
     */
    lasttab = -1;
}

/* The main event loop */
static void EventLoop(void)
{
    Word err;
    EventType e;
    long eventTimeout;
    DateType dtnow;
    Boolean onmainform=false;
    UInt cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    short i;

    ignorependown = false;
    eventTimeout = 1*tps;
    do {
        EvtGetEvent(&e, eventTimeout);

        /*
         * There is a funky find bug that gremlins always catches.  It is
         * a documented error, and so I don't let the find button be pressed
         * if we are running in gremlins.
         *
         * Also I have a notes popup form that is not aligned at the bottom
         * of the screen.  On occasion when returning from the launcher popup
         * gremlins will complain about an invalid window position.  This is
         * true, except that it is only a warning because it is inefficient.
         * This only happens during a redraw when coming back from the launcher
         * popup on PalmOS2.
         */

/*
        if (inposer) {
            if ((e.eType == keyDownEvent) &&
                (e.data.keyDown.modifiers & commandKeyMask) &&
                (e.data.keyDown.chr == launchChr)) continue;

            if ((e.eType == keyDownEvent) &&
                (e.data.keyDown.modifiers & commandKeyMask) &&
                (e.data.keyDown.chr == findChr)) continue;
        }
*/

        if (! SysHandleEvent(&e))
            if (! MenuHandleEvent(NULL, &e, &err))
                if (! ApplicationHandleEvent(&e))
                    FrmDispatchEvent(&e);

        if (e.eType == winEnterEvent) {
            if (e.data.winEnter.enterWindow ==
                (WinHandle) FrmGetFormPtr(MainForm))
                onmainform = true; else onmainform = false;
        }

        /*
         * Since you can press one of the 4 hard buttons at any time, I have
         * to test for them here, rather than inside of a form handler.
         *
         * I'm going to ignore this if the unit is just turning on.  This is
         * because I LIKE the ability to use a hard button to toggle between
         * two apps, but I also like to press a hard button to power on rather
         * that pushing the power button.  If I don't ignore the hard button
         * launch when we are turning on, then I can't leave my program running
         * and expect to turn the unit on without it launching the other app.
         */
        if ((e.eType == keyDownEvent) &&
            (e.data.keyDown.modifiers & commandKeyMask) &&
            !(e.data.keyDown.modifiers & poweredOnKeyMask)) {
            i = -1;
            switch (e.data.keyDown.chr) {
            case hard1Chr:
                if (PrefsR->LaunchPrefs.hardlaunchflag[0] == '1') i = 0;
                break;
            case hard2Chr:
                if (PrefsR->LaunchPrefs.hardlaunchflag[1] == '1') i = 1;
                break;
            case hard3Chr:
                if (PrefsR->LaunchPrefs.hardlaunchflag[2] == '1') i = 2;
                break;
            case hard4Chr:
                if (PrefsR->LaunchPrefs.hardlaunchflag[3] == '1') i = 3;
                break;
            }
            if ((i >= 0) && (DmGetNextDatabaseByTypeCreator(true, &searchInfo,
                sysFileTApplication, Prefshardlaunch[i], true, &cardNo,
                &dbID) == 0)) {
                SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdNormalLaunch, NULL);
            }
        }

        if (onmainform && (e.eType == keyDownEvent) &&
            (((e.data.keyDown.modifiers & commandKeyMask) &&
            ((e.data.keyDown.chr == hardPowerChr) ||
            (e.data.keyDown.chr == 0x11e))) || /* OS3.3 */
            (e.data.keyDown.modifiers & poweredOnKeyMask))) {
            DateSecondsToDate(TimGetSeconds(), &dtnow);
            if (DateToInt(dtnow) != DateToInt(CurrentDate)) {
                /*
                 * The device has been turned on and the day is different than
                 * when we left, so I need to refresh the current view.
                 */
                DateSecondsToDate(TimGetSeconds(), &CurrentDate);
                refreshlist(-1);
            }
            /* always do this at power on */
            DisplayFooter(false);
        }

    } while (e.eType != appStopEvent);

}

void DisplayFooter(Boolean force)
{
    DateType CurrentDate;
    static DateType oldDate={0, 0, 0};
    TimeType CurrentTime;
    static TimeType oldTime={0, 0};
    DateTimeType dt;
    UInt battv;
    static UInt oldbattv=0;
    RectangleType r;
    short i1, i2;
    FontID currFont=0;
    static Boolean iconsdrawn=false;
    FormPtr frm;
    VoidHand iconh;
    BitmapPtr iconp;

    r.topLeft.x = 0;
    r.topLeft.y = 137;
    r.extent.x = 90;
    r.extent.y = FntCharHeight();

    if (force) {
        WinDrawLine(r.topLeft.x, r.topLeft.y, 159, r.topLeft.y);
    }
/* Need more room so make it flush left
    r.topLeft.x += 2;
    r.extent.x -= 4;
*/
    r.topLeft.y += 1;

    /*
     * Update the date on the screen
     */
    DateSecondsToDate(TimGetSeconds(), &CurrentDate);
    if (force || (DateToInt(CurrentDate) != DateToInt(oldDate))) {
        if (!currFont) currFont = FntSetFont(boldFont);
        /* might as well use AppErrStr as a buffer */
        DateToDOWDMFormat(CurrentDate.month, CurrentDate.day,
            CurrentDate.year+1904, Prefsdfmt1, AppErrStr);
        WinEraseRectangle(&r, 0);
           DrawChars(AppErrStr, StrLen(AppErrStr), r.topLeft.x, r.topLeft.y);
        oldDate = CurrentDate;
    }
    r.topLeft.y += r.extent.y;

    /*
     * Update the time and battery voltage on the screen
     */
    TimSecondsToDateTime(TimGetSeconds(), &dt);
    CurrentTime.hours = dt.hour;
    CurrentTime.minutes = dt.minute;
    #if SDK_VERSION < 30
    battv = SysBatteryInfo(false, NULL, NULL, NULL, NULL, NULL);
    #else
    battv = SysBatteryInfo(false, NULL, NULL, NULL, NULL, NULL, NULL);
    #endif
    if (force || (TimeToInt(CurrentTime) != TimeToInt(oldTime)) ||
        (battv != oldbattv)) {
        if (!currFont) currFont = FntSetFont(boldFont);
        WinEraseRectangle(&r, 0);
        MenuEraseStatus(NULL);
        TimeToAscii(CurrentTime.hours, CurrentTime.minutes, Prefstfmt,
            AppErrStr);
           r.topLeft.x += DrawChars(AppErrStr, StrLen(AppErrStr), r.topLeft.x,
            r.topLeft.y)+4;
        oldTime = CurrentTime;
        i1 = numtotxt(battv, &AppErrStr[1], false)+1;
        AppErrStr[0] = AppErrStr[1];
        AppErrStr[1] = '.';
        AppErrStr[i1++] = 'v';
        i2 = FntCharsWidth(AppErrStr, i1)+2;
        WinDrawChars(AppErrStr, i1, r.topLeft.x, r.topLeft.y);
        oldbattv = battv;
    }

    /*
     * Draw the bitmap of a phone
     */
    if (force || !iconsdrawn) {
        if ((frm = FrmGetFormPtr(MainForm))) {
            /* clock */
            FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_clock), &r);
            if ((iconh = DmGetResource(bitmapRsc, bitmapID_clock))) {
                if ((iconp = MemHandleLock(iconh))) {
                    WinDrawBitmap(iconp, r.topLeft.x, r.topLeft.y);
                    MemHandleUnlock(iconh);
                }
                DmReleaseResource(iconh);
            }
            /* phone */
            FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_phone), &r);
            if ((iconh = DmGetResource(bitmapRsc, bitmapID_phone))) {
                if ((iconp = MemHandleLock(iconh))) {
                    WinDrawBitmap(iconp, r.topLeft.x, r.topLeft.y);
                    MemHandleUnlock(iconh);
                }
                DmReleaseResource(iconh);
            }
        }
        iconsdrawn = true;
    }

    if (currFont) FntSetFont(currFont);
}

short DrawChars(char *str, Word len, SWord x, SWord y)
{
    WinDrawChars(str, len, x, y);
    return FntCharsWidth(str, len);
}

/*
 * txt must be at least char[9]
 */
short numtotxt(long num, char *txt, Boolean leadingzeroes)
{
    long i1, i2, base;
    i2 = 0;
    base = 10000000;
    while (base) {
        if ((i1 = num/base) || i2 || (base == 1) || leadingzeroes)
            txt[i2++] = (char) (i1+48);
        num -= i1*base;
        base /= 10;
    }
    txt[i2] = (char) NULL;
    return i2;
}

Boolean OpenDatabases(void)
{
    Err err;
    Ptr p, p2;
    UInt mode=0;
    UInt cardNo;
    LocalID dbID, appInfoID;
    UInt attributes;
    DWord cdate, mdate;
    ULong creator;
    Boolean migrated=false;
    AddrAppInfoPtr appInfoPtr;
        DatabaseHdrType hdr;

    /*
     * Set these to zero in case any of the opens fail.
     */
    MainDB = 0;
    DatebookDB = 0;
    ToDoDB = 0;
    AddressDB = 0;
    PrefsRecHandle = 0;

    /* Determine if secret records should be shown. */
    if (!sysPrefs.hideSecretRecords) mode |= dmModeShowSecret;

    /* Open DatebookDB */
    if ((DatebookDB = DmOpenDatabaseByTypeCreator(DatebookDBType,
        DatebookAppID, mode | dmModeReadOnly))) {
        if (DmOpenDatabaseInfo(DatebookDB, &dbID, NULL, NULL, &cardNo, NULL)
            == 0) {
            if (DmDatabaseInfo(cardNo, dbID, hdr.name, NULL, NULL, &cdate,
                &mdate, NULL, NULL, &appInfoID, NULL, NULL, &creator) == 0) {
                dbcdate = cdate;
                dbmdate = mdate;
            }
        }
    }
    /* Open ToDoDB */
    if ((ToDoDB = DmOpenDatabaseByTypeCreator(ToDoDBType,
        ToDoAppID, mode | dmModeReadOnly))) {
        if (DmOpenDatabaseInfo(ToDoDB, &dbID, NULL, NULL, &cardNo, NULL)
            == 0) {
            if (DmDatabaseInfo(cardNo, dbID, hdr.name, NULL, NULL, &cdate,
                &mdate, NULL, NULL, &appInfoID, NULL, NULL, &creator) == 0) {
                tdcdate = cdate;
                tdmdate = mdate;
            }
        }
    }
    /* Open AddressDB */
    if ((AddressDB = DmOpenDatabaseByTypeCreator(AddressDBType,
        AddressAppID, mode | dmModeReadOnly))) {
        if (DmOpenDatabaseInfo(AddressDB, &dbID, NULL, NULL, &cardNo, NULL)
            == 0) {
            if (DmDatabaseInfo(cardNo, dbID, hdr.name, NULL, NULL, &cdate,
                &mdate, NULL, NULL, &appInfoID, NULL, NULL, &creator) == 0) {
                /*
                 * We want to get the sort order for the address book.  This
                 * controls how to display the contacts list.
                 */
                if (appInfoID) {
                    if ((appInfoPtr = (AddrAppInfoPtr)
                        MemLocalIDToLockedPtr(appInfoID, cardNo))) {
                        SortByCompany = appInfoPtr->misc.sortByCompany;
                        MemPtrUnlock(appInfoPtr);
                    }
                }
            }
        }
    }
    /* Open hotdateDB */
    MainDB = DmOpenDatabaseByTypeCreator(MainDBType, MainAppID,
        dmModeReadWrite);

    if (!MainDB) {
        /*
         * Create our database if it doesn't exist yet
         */
        if (DmCreateDatabase(0, MainDBName, MainAppID, MainDBType, false))
            return false;
        MainDB = DmOpenDatabaseByTypeCreator(MainDBType, MainAppID,
            dmModeReadWrite);
        if (!MainDB) return false;
        /* Set the backup bit.  This is to aid syncs with non Palm software. */
        DmOpenDatabaseInfo(MainDB, &dbID, NULL, NULL, &cardNo, NULL);
        DmDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL);
        attributes |= dmHdrAttrBackup;
        DmSetDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    }
    PrefsRecIndex = 0;
    if ((PrefsRecHandle = DmQueryRecord(MainDB, PrefsRecIndex)) != 0) {
        /*
         * A Prefs record already exists so we will do a sanity check.
         */
        if (MemHandleSize(PrefsRecHandle) >= sizeof(PrefsR->version)) {
            /*
             * Let's compare the version number.  If it doesn't match we will
             * try and migrate it to the new version.
             *
             * BTW, the current Prefs record version is PREFSVERSION.
             * Change this each time the Prefs record layout changes.
             */
            p = MemHandleLock(PrefsRecHandle);
            if (*p != (char) PREFSVERSION) {
                if (*p == (char) (PREFSVERSION-1)) {
                    MigratePrefs(PrefsR, (struct s126PrefsR *) p);
                    migrated = true;
                }
                MemPtrUnlock(p);
                err = DmRemoveRecord(MainDB, PrefsRecIndex);
                ErrFatalDisplayIf(err, ErrToString(err));
                PrefsRecHandle = 0;
            } else MemPtrUnlock(p);
        } else {
            err = DmRemoveRecord(MainDB, PrefsRecIndex);
            ErrFatalDisplayIf(err, ErrToString(err));
            PrefsRecHandle = 0;
        }
    }
    if (PrefsRecHandle == 0) {
        /*
         * Create the preferences record if it doesn't exist and
         * preload it with default values.
         */
        PrefsRecHandle = DmNewRecord(MainDB, &PrefsRecIndex,
            sizeof(struct sPrefsR));
        ErrFatalDisplayIf(!PrefsRecHandle,
            "Could not create prefs record.");
        p = MemHandleLock(PrefsRecHandle);
        if (migrated) p2 = (Ptr) PrefsR; else p2 = (Ptr) &DefaultPrefsR;
        err = DmWrite(p, 0, p2, sizeof(struct sPrefsR));
        ErrFatalDisplayIf(err, "Could not write default prefs.");
        MemPtrUnlock(p);
    }

    /*
     * Read in the preferences from the private database.  This also loads
     * some global variables with values for convenience.
     */
    ReadPrefsRec();

    /*
     * Check if the other databases opened correctly.  Abort if not.
     */
    if (!DatebookDB || !ToDoDB || !AddressDB) return false;

    return true;
}

void CloseDatabases(void)
{
    if (DatebookDB) DmCloseDatabase(DatebookDB);
    if (ToDoDB) DmCloseDatabase(ToDoDB);
    if (AddressDB) DmCloseDatabase(AddressDB);
    if (MainDB) {
        if (PrefsRecHandle) {
            if (PrefsRdirty) WritePrefsRec();
            DmReleaseRecord(MainDB, PrefsRecIndex, PrefsWeredirty);
            /*
             * Unlock the PrefsR structure
             */
            MemPtrUnlock(PrefsR);
        }
        DmCloseDatabase(MainDB);
    }
}

static void AppsPopupDrawItem(UInt itemNum, RectanglePtr bounds,
    CharPtr *unusedP)
{
    SysDBListItemType *listP;
        DatabaseHdrType hdr;
    SWord tl, tw;
    Boolean ignored;
    char ellipsisChar = horizEllipsisChr;

    CALLBACK_PROLOGUE

    if (AppsPopupH != NULL) {
        listP = MemHandleLock((VoidHand) AppsPopupH);
        if (DmDatabaseInfo(listP[itemNum].cardNo, listP[itemNum].dbID,
            hdr.name, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
            &hdr.creator) == 0) {
            tl = StrLen(hdr.name);
            tw = FntCharsWidth(hdr.name, tl);
            if (tw <= bounds->extent.x) {
                WinDrawChars(hdr.name, tl, bounds->topLeft.x,
                    bounds->topLeft.y);
            } else {
                tw = bounds->extent.x-FntCharWidth(ellipsisChar);
                FntCharsInWidth(hdr.name, &tw, &tl, &ignored);
                WinDrawChars(hdr.name, tl, bounds->topLeft.x,
                    bounds->topLeft.y);
                WinDrawChars(&ellipsisChar, 1, bounds->topLeft.x+tw,
                    bounds->topLeft.y);
            }
        }
        MemPtrUnlock(listP);
    }

    CALLBACK_EPILOGUE
}

static ULong AppsPopup(Word listid, Word triggerid, ULong curcreator)
{
    Word error;
    Int curSelection=-1;
    Int newSelection;
    Int i1, width, height;
    ULong newcreator=curcreator;
    ListPtr lst;
    ControlPtr trig;
        FormPtr frm;
    CharPtr label;
    RectangleType r;
    SysDBListItemType *listP;
        DatabaseHdrType hdr;
    WinHandle saved=0;
    VoidHand strH;
    CharPtr strP;

    /*
     * This is going to take a while, so we'll put a little message up
     * for the user.
     */
    if ((strH = DmGetResource(strRsc, strID_pleasewait))) {
        strP = MemHandleLock(strH);
        FntSetFont(boldFont);
        i1 = StrLen(strP);
        width = FntCharsWidth(strP, i1);
        height = FntCharHeight();
        r.topLeft.x = 80-((width+12)/2);
        r.topLeft.y = 80-((height+8)/2);
        r.extent.x = width+12;
        r.extent.y = height+8;
        if ((saved = WinSaveBits(&r, &error))) {
            WinEraseRectangle(&r, 0);
            r.topLeft.x++;
            r.topLeft.y++;
            r.extent.x -= 3;
            r.extent.y -= 3;
            WinDrawRectangleFrame(popupFrame, &r);
            WinDrawChars(strP, i1, r.topLeft.x+(r.extent.x/2)-(width/2),
                r.topLeft.y+(r.extent.y/2)-(height/2));
        }
        MemHandleUnlock(strH);
        DmReleaseResource(strH);
        FntSetFont(stdFont);
    }

    if (!SysCreateDataBaseList(sysFileTApplication, 0, &AppsPopupCount,
        &AppsPopupH, true)) {
        AppsPopupCount = 0;
        AppsPopupH = 0;
    } else {
        /* Figure out the current selection by comparing the creator ID's */
        listP = MemHandleLock((VoidHand) AppsPopupH);
        for (i1=0; i1 < AppsPopupCount; i1++) {
            if (listP[i1].creator == curcreator) {
                curSelection = i1;
                break;
            }
        }
        MemPtrUnlock(listP);
    }

    /* Get rid of the message */
    if (saved) WinRestoreBits(saved, --r.topLeft.x, --r.topLeft.y);

    if (AppsPopupCount) {
        frm = FrmGetActiveForm();
        lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listid));
        LstSetDrawFunction(lst, (ListDrawDataFuncPtr) AppsPopupDrawItem);

        LstSetListChoices(lst, NULL, AppsPopupCount);
        LstSetSelection(lst, curSelection);
        LstSetTopItem(lst, curSelection);

        newSelection = LstPopupList(lst);

        if ((newSelection != curSelection) && (newSelection != -1)) {
            listP = MemHandleLock((VoidHand) AppsPopupH);
            if (DmDatabaseInfo(listP[newSelection].cardNo,
                listP[newSelection].dbID, hdr.name, NULL, NULL, NULL, NULL,
                NULL, NULL, NULL, NULL, NULL, &hdr.creator) == 0) {
                newcreator = hdr.creator;
                /*
                 * Copy the name into the popup trigger label.
                 */
                trig = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, triggerid));
                label = CtlGetLabel(trig);
                StrCopy(label, hdr.name);
                CtlSetLabel(trig, label);
            }
            MemPtrUnlock(listP);
        }
    }

    if (AppsPopupH) MemHandleFree((VoidHand) AppsPopupH);

    return newcreator;
}

static void CheckPrefsR(void)
{
    /*
     * Do some sanity checks and rewrite the numeric fields to get
     * rid of whitespace or extraneous characters.  Maybe later on
     * I can figure out how to validate this in the input fields.
     */
}

static void ReadPrefsRec(void)
{
    Ptr p;
    short i;
    DWord val;

    if (MainDB) {
        p = MemHandleLock(PrefsRecHandle);
        MemMove(PrefsR, p, sizeof(struct sPrefsR));
        MemPtrUnlock(p);
        PrefsRdirty = false;
        PrefsWeredirty = false;
        /*
         * Here are some items that we make numerical equivalents of
         * for convenience.
         */
        switch (PrefsR->ApptPrefs.days) {
        case '0': Prefsdays = 7; break;
        case '1': Prefsdays = 14; break;
        case '2': Prefsdays = 21; break;
        case '3': Prefsdays = 31; break;
        default: Prefsdays = 7; break;
        }
        Prefstfmt = FindTimeFormat(PrefsR->ApptPrefs.timeformat);
        Prefsdfmt1 = FindDateFormat1(PrefsR->ApptPrefs.dateformat1);
        Prefsdfmt2 = FindDateFormat2(PrefsR->ApptPrefs.dateformat2);
        if ((apptToDoCategory =
            CategoryFind(ToDoDB, PrefsR->todocategory)) == dmAllCategories) {
            /*
             * This could be All or maybe the category has been deleted or
             * renamed.  So we'll go ahead and get the category name again
             * in this case and load it into our prefs.
             */
            MemSet(PrefsR->todocategory, sizeof(PrefsR->todocategory), 0);
            CategoryGetName(ToDoDB, apptToDoCategory, PrefsR->todocategory);
        }
        MemMove(&Prefsapptlaunch, PrefsR->LaunchPrefs.apptlaunch,
            sizeof(Prefsapptlaunch));
        MemMove(&Prefstodolaunch, PrefsR->LaunchPrefs.todolaunch,
            sizeof(Prefstodolaunch));
        MemMove(&Prefslowerleftlaunch, PrefsR->LaunchPrefs.lowerleftlaunch,
            sizeof(Prefslowerleftlaunch));
        MemMove(&Prefscontactslaunch, PrefsR->LaunchPrefs.contactslaunch,
            sizeof(Prefscontactslaunch));
        for (i=0; i < 4; i++) {
            MemMove(&Prefshardlaunch[i], PrefsR->LaunchPrefs.hardlaunch[i],
                sizeof(Prefshardlaunch[i]));
        }

        /*
         * Invalidate the cache if the secret pref has changed since the
         * last time we were run.
         */
        if ((FtrGet(MainAppID, 0, &val) != 0) ||
            ((val == 0) != (sysPrefs.hideSecretRecords))) {
            PrefsR->DayCache.listcount[0] = (char) 0;
            PrefsR->DayCache.listcount[1] = (char) 0;
            val = sysPrefs.hideSecretRecords ? 0:1;
            FtrSet(MainAppID, 0, val);
        }

    }
    CheckPrefsR();
}

static void WritePrefsRec(void)
{
    Ptr p;
    Err err;
    CheckPrefsR();
    if (MainDB) {
        p = MemHandleLock(PrefsRecHandle);
        err = DmWrite(p, 0, PrefsR, sizeof(struct sPrefsR));
        ErrFatalDisplayIf(err, "Could not write prefs.");
        MemPtrUnlock(p);
        /*
         * This is set until we exit the app at which time the ReleaseRecord()
         * is called.
         */
        PrefsWeredirty = true;
        PrefsRdirty = false;
        /* invalidate the cache because we probably just changed a pref */
        PrefsR->DayCache.listcount[0] = (char) 0;
        PrefsR->DayCache.listcount[1] = (char) 0;
    }
}

static short FindDateFormat1(char format)
{
    short fmt;
    switch (format) {
    case 'A': fmt = dfMDYWithSlashes; break;
    case 'B': fmt = dfDMYWithSlashes; break;
    case 'C': fmt = dfDMYWithDots; break;
    case 'D': fmt = dfDMYWithDashes; break;
    case 'E': fmt = dfYMDWithSlashes; break;
    case 'F': fmt = dfYMDWithDots; break;
    case 'G': fmt = dfYMDWithDashes; break;
    case 'H': fmt = dfMDYLongWithComma; break;
    case 'I': fmt = dfDMYLong; break;
    case 'J': fmt = dfDMYLongWithDot; break;
    case 'K': fmt = dfDMYLongWithComma; break;
    case 'L': fmt = dfYMDLongWithDot; break;
    case 'M': fmt = dfYMDLongWithSpace; break;
    default: fmt = dfMDYLongWithComma; break;
    }
    return fmt;
}

static short FindDateFormat2(char format)
{
    short fmt;
    switch (format) {
    case '0': fmt = dfMDYWithSlashes; break;
    case '1': fmt = dfDMYWithSlashes; break;
    case '2': fmt = dfDMYWithDots; break;
    case '3': fmt = dfDMYWithDashes; break;
    case '4': fmt = dfYMDWithDots; break;
    case '5': fmt = dfYMDWithDashes; break;
    case '6': fmt = dfYMDWithSlashes; break;
    default: fmt = dfMDYWithSlashes; break;
    }
    return fmt;
}

static short FindTimeFormat(char format)
{
    short fmt;
    switch (format) {
    case '0': fmt = tfColonAMPM; break;
    case '1': fmt = tfColon24h; break;
    case '2': fmt = tfDotAMPM; break;
    case '3': fmt = tfDot24h; break;
    default: fmt = tfColonAMPM; break;
    }
    return fmt;
}

static void LoadApptPrefsFields(void)
{
    FormPtr frm;
        ControlPtr ctl;
    ListPtr list;

    if ((frm = FrmGetFormPtr(ApptPrefsForm)) == 0) return;
    /* Default tab */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        listID_apptprefsdefaulttab));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_apptprefsdefaulttab));
    LstSetSelection(list, (short) (PrefsR->ApptPrefs.defaulttab-'0'));
    if (PrefsR->ApptPrefs.defaulttab < '0') PrefsR->ApptPrefs.defaulttab = '0';
    CtlSetLabel(ctl,
        LstGetSelectionText(list, (short) (PrefsR->ApptPrefs.defaulttab-'0')));
    /* Week (how many days?) */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        listID_apptprefsdays));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_apptprefsdays));
    LstSetSelection(list, (short) (PrefsR->ApptPrefs.days-'0'));
    if (PrefsR->ApptPrefs.days < '0') PrefsR->ApptPrefs.days = '0';
    CtlSetLabel(ctl, LstGetSelectionText(list,
        (short) (PrefsR->ApptPrefs.days-'0')));
    /* Week always starts on Monday */
    if (PrefsR->ApptPrefs.weekstart == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefsweekstart)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefsweekstart)), 0);
    /* Show date on week view */
    if (PrefsR->ApptPrefs.showdate == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefsshowdate)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefsshowdate)), 0);
    /* Show done items from datebk3 */
    if (PrefsR->ApptPrefs.showdone == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefsshowdone)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefsshowdone)), 0);
    /* Time format */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        listID_apptprefstimeformat));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_apptprefstimeformat));
    LstSetSelection(list, (short) (PrefsR->ApptPrefs.timeformat-'0'));
    if (PrefsR->ApptPrefs.timeformat < '0') PrefsR->ApptPrefs.timeformat = '0';
    CtlSetLabel(ctl,
        LstGetSelectionText(list, (short) (PrefsR->ApptPrefs.timeformat-'0')));
    /* Date at bottom */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        listID_apptprefsdateformat1));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_apptprefsdateformat1));
    LstSetSelection(list, (short) (PrefsR->ApptPrefs.dateformat1-'A'));
    if (PrefsR->ApptPrefs.dateformat1 < 'A')
        PrefsR->ApptPrefs.dateformat1 = 'A';
    CtlSetLabel(ctl,
        LstGetSelectionText(list, (short) (PrefsR->ApptPrefs.dateformat1-'A')));
    /* Date in list */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        listID_apptprefsdateformat2));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_apptprefsdateformat2));
    LstSetSelection(list, (short) (PrefsR->ApptPrefs.dateformat2-'0'));
    if (PrefsR->ApptPrefs.dateformat2 < '0')
        PrefsR->ApptPrefs.dateformat2 = '0';
    CtlSetLabel(ctl,
        LstGetSelectionText(list, (short) (PrefsR->ApptPrefs.dateformat2-'0')));
    /* Symbols: Note */
    if (PrefsR->ApptPrefs.symnote == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefssymnote)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefssymnote)), 0);
    /* Symbols: Alarm */
    if (PrefsR->ApptPrefs.symalarm == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefssymalarm)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefssymalarm)), 0);
    /* Symbols: Repeat */
    if (PrefsR->ApptPrefs.symrepeat == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefssymrepeat)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefssymrepeat)), 0);
}

static void UnloadApptPrefsFields(Boolean save)
{
    FormPtr frm;
    if ((frm = FrmGetFormPtr(ApptPrefsForm)) == 0) return;
    if (save) {
        /* Default tab */
        PrefsR->ApptPrefs.defaulttab =
            (char) (LstGetSelection(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_apptprefsdefaulttab)))+'0');
        /* Week (how many days?) */
        PrefsR->ApptPrefs.days = (char) (LstGetSelection(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_apptprefsdays)))+'0');
        switch (PrefsR->ApptPrefs.days) {
        case '0': Prefsdays = 7; break;
        case '1': Prefsdays = 14; break;
        case '2': Prefsdays = 21; break;
        case '3': Prefsdays = 31; break;
        default: Prefsdays = 7; break;
        }
        /* Week always starts on Monday */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefsweekstart))) == 1)
            PrefsR->ApptPrefs.weekstart = '1';
        else PrefsR->ApptPrefs.weekstart = '0';
        /* Show date on week view */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefsshowdate))) == 1)
            PrefsR->ApptPrefs.showdate = '1';
        else PrefsR->ApptPrefs.showdate = '0';
        /* Show done items from datebk3 */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefsshowdone))) == 1)
            PrefsR->ApptPrefs.showdone = '1';
        else PrefsR->ApptPrefs.showdone = '0';
        /* Time format */
        PrefsR->ApptPrefs.timeformat =
            (char) (LstGetSelection(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_apptprefstimeformat)))+'0');
        Prefstfmt = FindTimeFormat(PrefsR->ApptPrefs.timeformat);
        /* Date at bottom */
        PrefsR->ApptPrefs.dateformat1 =
            (char) (LstGetSelection(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_apptprefsdateformat1)))+'A');
        Prefsdfmt1 = FindDateFormat1(PrefsR->ApptPrefs.dateformat1);
        /* Date in list */
        PrefsR->ApptPrefs.dateformat2 =
            (char) (LstGetSelection(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_apptprefsdateformat2)))+'0');
        Prefsdfmt2 = FindDateFormat2(PrefsR->ApptPrefs.dateformat2);
        /* Symbols: Note */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefssymnote))) == 1)
            PrefsR->ApptPrefs.symnote = '1';
        else PrefsR->ApptPrefs.symnote = '0';
        /* Symbols: Alarm */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefssymalarm))) == 1)
            PrefsR->ApptPrefs.symalarm = '1';
        else PrefsR->ApptPrefs.symalarm = '0';
        /* Symbols: Repeat */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_apptprefssymrepeat))) == 1)
            PrefsR->ApptPrefs.symrepeat = '1';
        else PrefsR->ApptPrefs.symrepeat = '0';
    }
}

static void LoadToDoPrefsFields(void)
{
    FormPtr frm;
        ControlPtr ctl;
    ListPtr list;

    if ((frm = FrmGetFormPtr(ToDoPrefsForm)) == 0) return;
    /* ToDo sort */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        listID_todoprefssort));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_todoprefssort));
    LstSetSelection(list, (short) (PrefsR->ToDoPrefs.sort-'0'));
    if (PrefsR->ToDoPrefs.sort < '0') PrefsR->ToDoPrefs.sort = '0';
    CtlSetLabel(ctl,
        LstGetSelectionText(list, (short) (PrefsR->ToDoPrefs.sort-'0')));
    /* Show undated ToDos */
    if (PrefsR->ToDoPrefs.nodate == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsnodate)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsnodate)), 0);
    /* Show completed ToDos */
    if (PrefsR->ToDoPrefs.done == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsdone)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsdone)), 0);
    /* Show past due ToDos */
    if (PrefsR->ToDoPrefs.pastdue == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefspastdue)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefspastdue)), 0);
    /* Go to top if category changes */
    if (PrefsR->ToDoPrefs.gotop == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsgotop)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsgotop)), 0);
    /* Item range */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        listID_todoprefsrange));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_todoprefsrange));
    LstSetSelection(list, (short) (PrefsR->ToDoPrefs.range-'0'));
    if (PrefsR->ToDoPrefs.range < '0') PrefsR->ToDoPrefs.range = '0';
    CtlSetLabel(ctl,
        LstGetSelectionText(list, (short) (PrefsR->ToDoPrefs.range-'0')));
    /* Show U.A. until day of event */
    if (PrefsR->ToDoPrefs.showua == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsshowua)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsshowua)), 0);
}

static void UnloadToDoPrefsFields(Boolean save)
{
    FormPtr frm;
    if ((frm = FrmGetFormPtr(ToDoPrefsForm)) == 0) return;
    if (save) {
        /* ToDo sort */
        PrefsR->ToDoPrefs.sort = (char) (LstGetSelection(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_todoprefssort)))+'0');
        /* Show undated ToDos */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsnodate))) == 1)
            PrefsR->ToDoPrefs.nodate = '1';
        else PrefsR->ToDoPrefs.nodate = '0';
        /* Show completed ToDos */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsdone))) == 1)
            PrefsR->ToDoPrefs.done = '1';
        else PrefsR->ToDoPrefs.done = '0';
        /* Show past due ToDos */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefspastdue))) == 1)
            PrefsR->ToDoPrefs.pastdue = '1';
        else PrefsR->ToDoPrefs.pastdue = '0';
        /* Go to top if category changes */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsgotop))) == 1)
            PrefsR->ToDoPrefs.gotop = '1';
        else PrefsR->ToDoPrefs.gotop = '0';
        /* Item range */
        PrefsR->ToDoPrefs.range = (char) (LstGetSelection(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_todoprefsrange)))+'0');
        /* Show U.A. until day of event */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_todoprefsshowua))) == 1)
            PrefsR->ToDoPrefs.showua = '1';
        else PrefsR->ToDoPrefs.showua = '0';
    }
}

static void LoadLaunchPrefsFields(void)
{
    FormPtr frm;
        ControlPtr ctl;
        CharPtr label;
    UInt cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
        DatabaseHdrType hdr;
    LocalID appInfoID, sortInfoID;
    ULong AppID;
    short i;

    if ((frm = FrmGetFormPtr(LaunchPrefsForm)) == 0) return;
    /* Appt launch */
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_launchprefsapptlaunch));
    label = CtlGetLabel(ctl);
    MemMove(&AppID, PrefsR->LaunchPrefs.apptlaunch, sizeof(AppID));
    Prefsapptlaunch = AppID;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, AppID, true, &cardNo, &dbID) == 0) {
        if (DmDatabaseInfo(cardNo, dbID, hdr.name, &hdr.attributes,
            &hdr.version, &hdr.creationDate, &hdr.modificationDate,
            &hdr.lastBackupDate, &hdr.modificationNumber,
            &appInfoID, &sortInfoID, &hdr.type, &hdr.creator) == 0) {
            StrCopy(label, hdr.name);
            CtlSetLabel(ctl, label);
        }
    }
    /* ToDo launch */
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_launchprefstodolaunch));
    label = CtlGetLabel(ctl);
    MemMove(&AppID, PrefsR->LaunchPrefs.todolaunch, sizeof(AppID));
    Prefstodolaunch = AppID;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, AppID, true, &cardNo, &dbID) == 0) {
        if (DmDatabaseInfo(cardNo, dbID, hdr.name, &hdr.attributes,
            &hdr.version, &hdr.creationDate, &hdr.modificationDate,
            &hdr.lastBackupDate, &hdr.modificationNumber,
            &appInfoID, &sortInfoID, &hdr.type, &hdr.creator) == 0) {
            StrCopy(label, hdr.name);
            CtlSetLabel(ctl, label);
        }
    }
    /* Lower-left launch */
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_launchprefslowerleftlaunch));
    label = CtlGetLabel(ctl);
    MemMove(&AppID, PrefsR->LaunchPrefs.lowerleftlaunch, sizeof(AppID));
    Prefslowerleftlaunch = AppID;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, AppID, true, &cardNo, &dbID) == 0) {
        if (DmDatabaseInfo(cardNo, dbID, hdr.name, &hdr.attributes,
            &hdr.version, &hdr.creationDate, &hdr.modificationDate,
            &hdr.lastBackupDate, &hdr.modificationNumber,
            &appInfoID, &sortInfoID, &hdr.type, &hdr.creator) == 0) {
            StrCopy(label, hdr.name);
            CtlSetLabel(ctl, label);
        }
    }
    /* Contacts launch */
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_launchprefscontactslaunch));
    label = CtlGetLabel(ctl);
    MemMove(&AppID, PrefsR->LaunchPrefs.contactslaunch, sizeof(AppID));
    Prefscontactslaunch = AppID;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, AppID, true, &cardNo, &dbID) == 0) {
        if (DmDatabaseInfo(cardNo, dbID, hdr.name, &hdr.attributes,
            &hdr.version, &hdr.creationDate, &hdr.modificationDate,
            &hdr.lastBackupDate, &hdr.modificationNumber,
            &appInfoID, &sortInfoID, &hdr.type, &hdr.creator) == 0) {
            StrCopy(label, hdr.name);
            CtlSetLabel(ctl, label);
        }
    }
    /* Hard button launches */
    for (i=0; i < 4; i++) {
        ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            popuptriggerID_launchprefshard1launch+i));
        label = CtlGetLabel(ctl);
        MemMove(&AppID, PrefsR->LaunchPrefs.hardlaunch[i], sizeof(AppID));
        Prefshardlaunch[i] = AppID;
        if (DmGetNextDatabaseByTypeCreator(true, &searchInfo,
            sysFileTApplication, AppID, true, &cardNo, &dbID) == 0) {
            if (DmDatabaseInfo(cardNo, dbID, hdr.name, &hdr.attributes,
                &hdr.version, &hdr.creationDate, &hdr.modificationDate,
                &hdr.lastBackupDate, &hdr.modificationNumber,
                &appInfoID, &sortInfoID, &hdr.type, &hdr.creator) == 0) {
                StrCopy(label, hdr.name);
                CtlSetLabel(ctl, label);
            }
        }
        if (PrefsR->LaunchPrefs.hardlaunchflag[i] == '1')
            CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkboxID_launchprefshard1launch+i)), 1);
        else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkboxID_launchprefshard1launch+i)), 0);
    }
}

static void UnloadLaunchPrefsFields(Boolean save)
{
    FormPtr frm;
    short i;

    if ((frm = FrmGetFormPtr(LaunchPrefsForm)) == 0) return;
    if (save) {
        /* Appt launch */
        MemMove(PrefsR->LaunchPrefs.apptlaunch, &Prefsapptlaunch,
            sizeof(Prefsapptlaunch));
        /* ToDo launch */
        MemMove(PrefsR->LaunchPrefs.todolaunch, &Prefstodolaunch,
            sizeof(Prefstodolaunch));
        /* Lower-left launch */
        MemMove(PrefsR->LaunchPrefs.lowerleftlaunch, &Prefslowerleftlaunch,
            sizeof(Prefslowerleftlaunch));
        /* Contacts launch */
        MemMove(PrefsR->LaunchPrefs.contactslaunch, &Prefscontactslaunch,
            sizeof(Prefscontactslaunch));
        /* Hard button launches */
        for (i=0; i < 4; i++) {
            MemMove(PrefsR->LaunchPrefs.hardlaunch[i], &Prefshardlaunch[i],
                sizeof(Prefshardlaunch[i]));
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkboxID_launchprefshard1launch+i))) == 1)
                PrefsR->LaunchPrefs.hardlaunchflag[i] = '1';
            else PrefsR->LaunchPrefs.hardlaunchflag[i] = '0';
        }
    } else {
        /* revert a few settings */
        MemMove(&Prefsapptlaunch, PrefsR->LaunchPrefs.apptlaunch,
            sizeof(Prefsapptlaunch));
        MemMove(&Prefstodolaunch, PrefsR->LaunchPrefs.todolaunch,
            sizeof(Prefstodolaunch));
        MemMove(&Prefslowerleftlaunch, PrefsR->LaunchPrefs.lowerleftlaunch,
            sizeof(Prefslowerleftlaunch));
        MemMove(&Prefscontactslaunch, PrefsR->LaunchPrefs.contactslaunch,
            sizeof(Prefscontactslaunch));
        for (i=0; i < 4; i++) {
            MemMove(&Prefshardlaunch[i], PrefsR->LaunchPrefs.hardlaunch[i],
                sizeof(Prefshardlaunch[i]));
        }
    }
}

/*
 * The program starts here.
 */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {
        err = StartApplication();
        if (err) return err;
        EventLoop();
        StopApplication();
    } else if (cmd == sysAppLaunchCmdAlarmTriggered) {
        /*
         * Play the pizza song...
         */
        AlarmTriggered((SysAlarmTriggeredParamType *)cmdPBP);
    } else if (cmd == sysAppLaunchCmdDisplayAlarm) {
        /*
         * This is for the pizza timer feature.
         */
        DisplayAlarm();
    } else {
        return sysErrParamErr;
    }

    return 0;
}

/*
 * Update information on week label given the provided arguments
 *
 * Added by FME - May 6, 2000. Will be called from MainFormHandleEvent
 */
static void UpdateWeekLabel(FormPtr frm, DateType BeginDate)
{
    ControlPtr ptr;

    /*
     * Create an appropriate MM/DD style label for the week tab.
     */
    DateToAscii(BeginDate.month, BeginDate.day, BeginDate.year+1904,
                Prefsdfmt2, AppErrStr);

    /* slash, dot, or dash */
    switch (Prefsdfmt2) {
    case dfYMDWithDots:
    case dfYMDWithDashes:
    case dfYMDWithSlashes:
        StrNCopy(weeklabel, &AppErrStr[3], StrLen(AppErrStr)-3);
        weeklabel[StrLen(AppErrStr)-3] = (char) NULL;
        break;
    default:
        StrNCopy(weeklabel, AppErrStr, StrLen(AppErrStr)-3);
        weeklabel[StrLen(AppErrStr)-3] = (char) NULL;
        break;
    }

    ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, pbuttonID_week));
    CtlSetLabel(ptr, weeklabel);

    MainFormLoadTable(frm, 1, 0, BeginDate, Prefsdays);
    ShowScrollArrows(frm);
}
