/* $Id: hotdate.h,v 1.32 2000/03/10 16:30:05 chrisf Exp $ */

/*
Hot Date - A DatebookDB displayer for the PalmPilot
Copyright (C) 1999 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __HOTDATE_H__
#define __HOTDATE_H__

#include "datebook.h"

#define PREFSVERSION 21

#define AppErrStrLen 40
#define ErrToString(err) SysErrString(err, AppErrStr, AppErrStrLen)

#ifndef UCASE
#define UCASE(_a) ((_a)>='a' && (_a)<='z' ? (_a)-32:(_a))
#endif
#ifndef	MAX
#define	MAX(_a,_b) ((_a)<(_b)?(_b):(_a))
#endif
#ifndef	MIN
#define	MIN(_a,_b) ((_a)<(_b)?(_a):(_b))
#endif

/* Ways to determine SDK header version (PalmOS2, 3 etc) */

#ifndef SDK_VERSION

/* 3.5 uses <PalmTypes.h> instead of <Common.h>.  */
#if defined(__PALMTYPES_H__)
#define SDK_VERSION 35

/* Otherwise, all versions of Common.h unconditionally #include
   <BuildRules.h>.  #defines corresponding to new functionality in
   each SDK were added to this file.  */

/* 3.1 added Japanese support.  */
#elif defined(LANGUAGE_JAPANESE)
#define SDK_VERSION 31

/* 3.0 introduced the v2 memory manager.  */
#elif defined(MEMORY_VERSION_2)
#define SDK_VERSION 30

/* 2.0 added Italian support.  */
#elif defined(LANGUAGE_ITALIAN)
#define SDK_VERSION 20

#else
#define SDK_VERSION 10
#endif

#endif

/* Database creators */

#define MainAppID 'HOTT'
#define MainDBType 'DATA'
#define MainDBName "hotdateDB"
#define DatebookAppID 'date'
#define DatebookDBType 'DATA'
#define DatebookDBName "DatebookDB"
#define ToDoAppID 'todo'
#define ToDoDBType 'DATA'
#define ToDoDBName "ToDoDB"
#define AddressAppID 'addr'
#define AddressDBType 'DATA'
#define AddressDBName "AddressDB"

#define MAXCONTACTS 32

/* Appt Preferences */
struct sApptPrefs {
    char defaulttab;
    char days;
    char weekstart;
    char showdate;
    char showdone;
    char timeformat;
    char dateformat1;
    char dateformat2;
    char symnote;
    char symalarm;
    char symrepeat;
};

/* ToDo Preferences */
struct sToDoPrefs {
    char sort;
    char nodate;
    char done;
    char pastdue;
    char gotop;
    char range;
    char showua;
};

/* Launch Preferences */
struct sLaunchPrefs {
    char apptlaunch[4];
    char todolaunch[4];
    char lowerleftlaunch[4];
    char contactslaunch[4];
    char hardlaunchflag[4];
    char hardlaunch[4][4];
};

/* Day cache */
struct sDayCache {
    char date[sizeof(DateType)]; /* the date the cache was created */
    char dbcdate[4], dbmdate[4]; /* DatebookDB create/modify dates */
    char tdcdate[4], tdmdate[4]; /* ToDoDB create/modify dates */
    char alarmline[2];
    char todoline[2];
    char listcount[2];
    char list[apptMaxPerDay*sizeof(LineItemType)];
};

/* Contacts */
struct sContacts {
    char listcount;
    char list[MAXCONTACTS*sizeof(ULong)]; /* AddressDB record uniqueIDs */
};

/*
 * A preferences record which consists of back-to-back
 * structures.  It also has a version char which starts at 1 to
 * allow testing for preferences migration in future versions.
 */
struct sPrefsR {
    char version;
    struct sApptPrefs ApptPrefs;
    struct sToDoPrefs ToDoPrefs;
    struct sLaunchPrefs LaunchPrefs;
    char todocategory[33];
    struct sDayCache DayCache;
    struct sContacts Contacts;
};

/*
 * THESE ARE THE STRUCTURES FOR THE PREVIOUS VERSION'S PREFERENCES.  I USE
 * THEM FOR MIGRATION PURPOSES.
 */

/*******************************************************/

/* Appt Preferences */
struct s126ApptPrefs {
    char defaulttab;
    char days;
    char weekstart;
    char showdate;
    char showdone;
    char timeformat;
    char dateformat1;
    char dateformat2;
    char symnote;
    char symalarm;
    char symrepeat;
};

/* ToDo Preferences */
struct s126ToDoPrefs {
    char sort;
    char nodate;
    char done;
    char pastdue;
    char gotop;
    char range;
    char showua;
};

/* Launch Preferences */
struct s126LaunchPrefs {
    char apptlaunch[4];
    char todolaunch[4];
    char lowerleftlaunch[4];
    char hardlaunchflag[4];
    char hardlaunch[4][4];
};

struct s126PrefsR {
    char version;
    struct s126ApptPrefs ApptPrefs;
    struct s126ToDoPrefs ToDoPrefs;
    struct s126LaunchPrefs LaunchPrefs;
    char todocategory[33];
};

/*******************************************************/

/*
 * Function prototypes
 */
void MainFormLoadTable(FormPtr frm, short redraw, UInt top, DateType date,
    UInt days);
void ShowScrollArrows(FormPtr frm);
void DisplayFooter(Boolean force);
short DrawChars(char *str, Word len, SWord x, SWord y);
short numtotxt(long num, char *txt, Boolean leadingzeroes);
Boolean OpenDatabases(void);
void CloseDatabases(void);
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags);

#endif
